/*
 *  Daubechies.cpp
 *  
 *
 *  Copyright 2014 INRIA. All rights reserved.
 *
 */
#include <iostream>
#include <fstream>
#include "Daubechies.h"
#include "math.h"
using namespace cimg_library;
using namespace std;

/**
 *  constructors
 */
Daubechies::Daubechies(): nMoments(-1){ };

Daubechies::Daubechies(int P, int type): nMoments(P), wavType(type) {
  
  CImg<double> tempFilter;
  int tempOff;
  
  //-first we get h filter
  //----------------------
  
  //-we also get filters' offset
  switch (wavType) {
    case 0:
      //-Daubechies
      setDaubechiesFilters();
      break;
    case 1:
      //-Coiflets
      setCoifletFilters();
      break;      
    default:
      //-use Daubechies as default
      wavType = 0;
      setDaubechiesFilters();
      break;
  }
  
  //-then we compute every other filters
  //------------------------------------
  
  //-g is the mirror of h, multiplying by (-1) every second value
  g.assign(h.get_mirror('x'));
  cimg_forX(g,k){
    g(k) *= pow(-1.,k+1);
  }
  //-and mirror filters
  hb = h.get_mirror('x');
  gb = g.get_mirror('x');  
  //--compute interpolation filter
  tempFilter = convolve(h,hOff,hb,hbOff,tempOff,0);
  hi = decimate(tempFilter);
  //--filter for transposed application
  hib = hi.get_mirror('x');
};

/**
 *  getters
 */
int Daubechies::get_nMoments() const{
  return nMoments;
};

int Daubechies::get_type() const{
  return wavType;
};

CImg<double> Daubechies::get_h() const{
  return h;
};

CImg<double> Daubechies::get_g() const{
  return g;
};

/**
 *  info
 */
void Daubechies::info(bool printFilters){


  switch (wavType) {
    case 0:
      cout << "Daubechies";  
      break;
    case 1:
      cout << "Coiflet";  
      break;      
    default:
      cout << "You should not read this no matter what ?!";
      break;
  }
  cout<<", "<<nMoments<<" vanishing moments."<<endl;
  if(printFilters){
    cout<<"<-------H-&-G-filters------>"<<endl;
    cimg_forX(h, k) {
      cout<<"["<<k+hOff<<"] "<<h(k)<<"\t["<<k+gOff<<"] "<<g(k)<<endl;
    }
    cout<<"<-------------------------->"<<endl;
  }
};

/*================================================*/
/* DECOMPOSITION / RECONSTRUCTION / INTERPOLATION */
/*================================================*/

/*----------------------*/
/* Full transformations */
/*----------------------*/

/**
 *  image2coeff1D
 */
CImgList<double> Daubechies::image2coeff1D(const CImg<double>& image, int level){
  
  return decompose1D(image/sqrt(image.width()),level);
};

/**
 *  image2coeff2D
 */
CImgList<double> Daubechies::image2coeff2D(const CImg<double>& image, int level){

  return decompose2D(image/sqrt(image.size()),level);
};

/**
 *  coeff2image1D
 */
CImg<double> Daubechies::coeff2image1D(const CImgList<double>& coeff, int level){
  
  return projection1D(coeff, coeff.size()-1-level);
};

/**
 *  coarseCoeff2image2D
 */
CImg<double> Daubechies::coarseCoeff2image2D(const CImgList<double>& coeff, int level, int detailLimit){

  CImg<double> coarse = compose2D(coeff,detailLimit), res;
  res = coarseCompose2D(coarse,level-detailLimit);
  return res*sqrt(res.size());
};

/**
 *  coeff2image2D
 */
CImg<double> Daubechies::coeff2image2D(const CImgList<double>& coeff, int level){
  
  return projection2D(coeff, coeff.size()-1-level);
};

/*--------------*/
/* Coefficients */
/*--------------*/

/**
 *  allocCoeffs
 */
CImgList<double> Daubechies::allocCoeffs(int width, int height, int &level){

  int n0w = (int)log2(width), n0h = (int)log2(height);
  CImgList<double> coeffs;
    
  //-check
  if ((n0w<level)||(n0h<level)) {
    cout << "\t /!\\ Too many decomposition requested (" << level <<"), corrected to " << min(n0w,n0h) << endl;
    level = min(n0w,n0h);
  }
  else if (level<0){
    cout << "\t /!\\ Impossible decomposition requested (" << level <<"), corrected to " << 0 << endl;
    level = 0;
  }
  coeffs.assign(level+1);
  
  //----allocation details coeffs : details are TRANSPOSED ! see decompose2D/compose2D
  for (int k=0; k<level; k++) {
    coeffs(level-k) = CImg<double>(height>>(k+1),width>>(k+1),3,1,0);
    //-3 slices per image => d1, d2 & d3.
  }
  //-coarse approximation : only one slice, NOT TRANSPOSED.
  coeffs(0) = CImg<double>(width>>level,height>>level,1,1,0); 
  return coeffs;
};

/**
 *  regroupCoeffPyramid
 */
CImg<double> Daubechies::regroupCoeffPyramid(const CImgList<double> &pyramid){
  
  CImg<double> result;
  //-dims are transposed within pyramids !
  int dh = 2*pyramid(pyramid.size()-1).width(), dw = 2*pyramid(pyramid.size()-1).height(), offx, offy;
  
  result.assign(dw,dh,1,1,0.);
  //-coarse
  cimg_forXY(pyramid(0),p,q){
    result(p,q) = pyramid(0)(p,q);
  }
  //-details (stored transposed within pyramid !)
  for (unsigned int s=1; s<pyramid.size(); s++) {
    offy = pyramid(s).width();
    offx = pyramid(s).height();
    cimg_forXY(pyramid(s),p,q){
      result(q+offx,p) = pyramid(s)(p,q,0);
    }
    cimg_forXY(pyramid(s),p,q){
      result(q,p+offy) = pyramid(s)(p,q,1);
    }  
    cimg_forXY(pyramid(s),p,q){
      result(q+offx,p+offy) = pyramid(s)(p,q,2);
    }
  }
  return result;
};

/*-------------*/
/* Filterbanks */
/*-------------*/

/**
 *  stepDown
 */
CImgList<double> Daubechies::stepDown(const CImg<double>& a){

  CImg<double> temp1, temp2;
  CImgList<double> res(2);
  int i1,i2;
  
  temp1.assign(convolve(a, 0, hb, hbOff, i1, 1));
  temp2.assign(convolve(a, 0, gb, gbOff, i2, 1));
  res(0).assign(decimate(temp1));
  res(1).assign(decimate(temp2));
  temp1.assign();
  temp2.assign();
  return res;
};

CImg<double> Daubechies::stepDownI(const CImg<double>& a){
  
  CImg<double> temp1, temp2;
  CImg<double> res(a,"x");
  int i1,i2, mid=a.width()/2;
  
  temp1.assign(convolve(a, 0, hb, hbOff, i1, 1));
  temp2.assign(convolve(a, 0, gb, gbOff, i2, 1));
  res.get_shared_points(0,mid-1) = decimate(temp1);
  res.get_shared_points(mid,res.width()-1) = decimate(temp2);
  return res;
};

/**
 *  stepUp
 */
CImg<double> Daubechies::stepUp(const CImg<double>& a, const int& Aoffset, 
                   const CImg<double>& b, const int& Boffset, 
                   int& Roffset, int border){
  
  CImg<double> temp0, temp1, temp2;
  int i1, i2;

  temp1.assign(convolve(expand(a), 2*Aoffset, h, hOff, i1, border));
  temp2.assign(convolve(expand(b), 2*Boffset, g, gOff, i2, border));
  temp0.assign(add(temp1, i1, temp2, i2, Roffset));
  temp1.assign();
  temp2.assign();             
  return temp0;
};

CImg<double> Daubechies::stepUpI(const CImg<double>& a,
                                const CImg<double>& b){
  
  CImg<double> temp1, temp2;
  int i1, i2;
  
  temp1.assign(convolve(expand(a), 0, h, hOff, i1, 1));
  temp2.assign(convolve(expand(b), 0, g, gOff, i2, 1));
  return add(temp1, 0, temp2, 0, i1);
};

/**
 *  coarseStepUp
 */
CImg<double> Daubechies::coarseStepUp(const CImg<double>& a, const int& Aoffset, 
                                      int& Roffset, int border){
  
  return convolve(expand(a), 2*Aoffset, h, hOff, Roffset, border);
};

/**
 *  decompose1D
 */
CImgList<double> Daubechies::decompose1D(const CImg<double>& signal, int level){

  CImgList<double> data(level+1);
  // level+1 images : level details + 1 coarse
  CImgList<double> temp(2);
  
  temp(0) = signal;
  for (int k=0; k<level; k++) {
    temp = stepDown(temp(0)); 
    data(level-k) = temp(1);
  }
  data(0) = temp(0);
  temp.assign();
  return data; 
};

CImg<double> Daubechies::decompose1DI(const CImg<double>& signal, int level){
  
  CImg<double> res(signal,false);
  //CImg<double> temp;
  int dim = signal.width();
  
  for (int k=0; k<level; k++) {
    dim = signal.width()>>k;
        res.get_shared_points(0,dim-1) =  stepDownI(res.get_shared_points(0,dim-1)); 
  }
  return res; 
};

CImg<double> Daubechies::decompose1DI(const CImg<double>& signal, int level, double zeta){
	CImg<double> res(signal,false);
	//CImg<double> temp;
	int dim = signal.width();
	//double tmp=0;
	for (int k=0; k<level; k++) {
		dim = signal.width()>>k;
 		//tmp+=zeta;
        res.get_shared_points(0,dim-1) =  stepDownI(res.get_shared_points(0,dim-1))*pow(2.,zeta); 
	}
	return res; 
};

/**
 *  compose1D
 */
CImg<double> Daubechies::compose1DI(const CImg<double>& data, int Clevel, double zeta){
	
	CImg<double> result(data,false);
	int level=(int)(log2((double)result.width()))-Clevel;
	//double tmp=0;
	int dim;
	for (int k=0; k<level; k++) {
		dim = result.width()>>k;
 		//tmp+=zeta;
        result.get_shared_points(0,dim-1) *=pow(2.,-1.*zeta); 
	}
	int cdim = 1<<Clevel;
	while (cdim < result.width()) {
		//tmp+=zeta;
		result.get_shared_points(0,2*cdim-1) = stepUpI(result.get_shared_points(0,cdim-1),result.get_shared_points(cdim,2*cdim-1));
		cdim *= 2;
	}
	
 	return result;
};
CImg<double> Daubechies::compose1D(const CImgList<double>& data, int& Roffset, int level, int border){

  CImg<double> result;
  int it;
  
  result = data(0);
  Roffset = 0;
  for (int k=0; k<level; k++) {
    result = stepUp(result,Roffset,data(k+1),0,it,border);
    Roffset = it;
  }
  return result;
};

CImg<double> Daubechies::compose1DI(const CImg<double>& data, int Clevel){
  
  CImg<double> result(data,false);
  int cdim;
  
  cdim = 1<<Clevel;
  while (cdim < result.width()) {
    result.get_shared_points(0,2*cdim-1) = stepUpI(result.get_shared_points(0,cdim-1),result.get_shared_points(cdim,2*cdim-1));
    cdim *= 2;
  }
  return result;
};

/**
 *  decompose2D
 */
CImgList<double> Daubechies::decompose2D(const CImg<double>& signal, int &level){

  CImgList<double> Data, temp(2);
  CImg<double> temp0, temp1, temp2;
  int currentW, currentH;
  
  //-----allocation coeffs
  Data.assign(allocCoeffs(signal.width(),signal.height(),level));
  
  //-----More Fire
  temp0.assign(signal);
  
  if (level>0) {
    //---we assume the image dims are powers of 2.
    currentW = signal.width();
    currentH = signal.height();
    for (int k=0; k<level; k++) {
      //--reset temps
      temp1.assign(currentW>>1,currentH,1,1,0);
      temp2.assign(currentW>>1,currentH,1,1,0);
      //--along rows
      for (int p=0; p<currentH; p++) {
        temp.assign(stepDown(temp0.get_shared_line(p)));
        temp1.get_shared_line(p) = temp(0);
        temp2.get_shared_line(p) = temp(1);
      }
      temp1.transpose();
      temp2.transpose();
      //--decrease width
      currentW >>= 1; 
      temp0.assign(currentH>>1,currentW,1,1,0);
      //--along columns
      for (int p=0; p<currentW; p++) {
        temp.assign(stepDown(temp1.get_shared_line(p)));
        temp0.get_shared_line(p) = temp(0);
        Data(level-k).get_shared_line(p,0) = temp(1);
        temp.assign(stepDown(temp2.get_shared_line(p)));
        Data(level-k).get_shared_line(p,1) = temp(0);
        Data(level-k).get_shared_line(p,2) = temp(1);
      }
      //--transpose back
      temp0.transpose();
      //--decrease height
      currentH >>= 1; 
    }
  }
  Data(0) = temp0;

  //-----return                        
  return Data;
};

/**
 *  compose2D
 */
CImg<double> Daubechies::compose2D(const CImgList<double>& data, int level){

  CImg<double> temp0, temp1, temp2, temp;
  int currentW, currentH, tOffset; 
  
  temp0.assign(data(0));
  currentW = temp0.width();
  currentH = temp0.height();
  //---More Fire
  for (int k=0; k<level; k++) {
    temp1.assign(currentH<<1,currentW,1,1,0);
    temp2.assign(currentH<<1,currentW,1,1,0);
    temp0.transpose();
    //--columns
    for (int p=0; p<currentW; p++) {
      temp1.get_shared_line(p) = stepUp(temp0.get_shared_line(p), 0, data(k+1).get_shared_line(p,0), 0, tOffset, 1);
      temp2.get_shared_line(p) = stepUp(data(k+1).get_shared_line(p,1), 0, data(k+1).get_shared_line(p,2), 0, tOffset, 1);
    }
    //--transpose back
    temp1.transpose();
    temp2.transpose();
    //-increase height
    currentH <<= 1;
    temp0.assign(currentW<<1,currentH,1,1,0);
    //--lines
    for (int p=0; p<currentH; p++) {
      temp0.get_shared_line(p) = stepUp(temp1.get_shared_line(p), 0, temp2.get_shared_line(p), 0, tOffset, 1);
    }
    //--increase width
    currentW <<= 1;
  }
  //-----return
  return temp0;
};

/**
 *  coarseCompose2D
 */
CImg<double> Daubechies::coarseCompose2D(const CImg<double>& coarse, int level){
  
  CImg<double> temp0, temp1;
  int currentW, currentH, tOffset; 
  
  temp0.assign(coarse);
  currentW = temp0.width();
  currentH = temp0.height();
  //---More Fire
  for (int k=0; k<level; k++) {
    temp1.assign(currentH<<1,currentW,1,1,0);
    temp0.transpose();
    //--columns
    for (int p=0; p<currentW; p++) {
      temp1.get_shared_line(p) = coarseStepUp(temp0.get_shared_line(p), 0, tOffset, 1);
    }
    //--transpose back
    temp1.transpose();
    //-increase height
    currentH <<= 1;
    temp0.assign(currentW<<1,currentH,1,1,0);
    //--lines
    for (int p=0; p<currentH; p++) {
      temp0.get_shared_line(p) = coarseStepUp(temp1.get_shared_line(p), 0, tOffset, 1);
    }
    //--increase width
    currentW <<= 1;
  }
  //-----return
  return temp0;
};

/**
 *  WIP - anisDecompose2D !!!!! square only !!!!!!
 */
CImg<double> Daubechies::anisDecompose2D(const CImg<double>& signal, int &level){
	return anisDecompose2D(signal,level,1);
}
CImg<double> Daubechies::anisDecompose2D(const CImg<double>& signal, int &level, bool normalize){

  CImg<double> result(signal,"xy",0.);
  
  //-lines
  cimg_forY(result,p){
    //-get 1d decomposition
    result.get_shared_line(p) = decompose1DI(signal.get_shared_line(p), level);
  }
  //temp.transpose();
  result.transpose();
  //-columns
  cimg_forY(result,p){
    //-get 1d decomposition
    result.get_shared_line(p) = decompose1DI(result.get_shared_line(p), level);
  }  
	if (normalize){result/=sqrt((double)(result.size()));}	
  return result;
};

CImg<double> Daubechies::anisDecompose2DFracNormalize(const CImg<double>& signal, int &level, double &zeta){
	
	
	CImg<double> result(signal,"xy",0.);
	//-lines
	cimg_forY(result,p){
		//-get 1d decomposition
		result.get_shared_line(p) = decompose1DI(signal.get_shared_line(p), level,zeta);
	}
	//temp.transpose();
	result.transpose();
	//-columns
	cimg_forY(result,p){
		//-get 1d decomposition
		result.get_shared_line(p) = decompose1DI(result.get_shared_line(p), level,zeta);
	}  
	return result/sqrt((double)(result.size()));
};


/**
 *  WIP - anisCompose2D !!!!! square only !!!!!!
 */


CImg<double> Daubechies::anisCompose2D(const CImg<double>& data, int &level, bool normalize){
    CImg<double> result(data,false);
    int coarseLvl;
    
    //-columns (because data is transposed)
    coarseLvl = (int)(log2((double)result.width())) - level;
    cimg_forY(result,p){
        //-get 1d composition
        result.get_shared_line(p) = compose1DI(result.get_shared_line(p), coarseLvl);
    }
    result.transpose();
    //-then lines
    coarseLvl = (int)(log2((double)result.width())) - level;
    cimg_forY(result,p){
        //-get 1d composition
        result.get_shared_line(p) = compose1DI(result.get_shared_line(p), coarseLvl);
    }
    if (normalize){result*=sqrt((double)(result.size()));}
    return result;

};

CImg<double> Daubechies::anisCompose2D(const CImg<double>& data, int &level){
  
   return anisCompose2D(data,level,1);
};

CImg<double> Daubechies::anisCompose2DFracNormalize(const CImg<double>& data, int &level, double &zeta){

	CImg<double> result(data,false);
	int coarseLvl;
	
	//-columns (because data is transposed)
	coarseLvl = (int)(log2((double)result.width())) - level;
	cimg_forY(result,p){
		//-get 1d composition
		result.get_shared_line(p) = compose1DI(result.get_shared_line(p), coarseLvl,zeta);
	}
	result.transpose();
	//-then lines
	coarseLvl = (int)(log2((double)result.width())) - level;
	cimg_forY(result,p){
		//-get 1d composition
		result.get_shared_line(p) = compose1DI(result.get_shared_line(p), coarseLvl,zeta);
	}
	return result*sqrt((double)(result.size()));
};

/*-----------------*/
/* Basis functions */
/*-----------------*/

/**
 *  wFunc
 */
CImg<double> Daubechies::wFunc(int& Roffset, int level, int func){
  
  CImg<double> A(1,1,1,1, 1.);
  CImg<double> B(1,1,1,1, 0.);
  CImg<double> result;
  CImgList<double> data(level+1);
  int N = 2<<level-1;
    
  for (int k=0; k<=level; k++) data(k) = B;
  
  switch (func) {
    case 0:
      //-Phi
      data(0) = A;
      break;
    case 1:
      //-Psi
      data(1) = A;
      break;
  }
  
  result = compose1D(data, Roffset, level, 0);
  return result*sqrt(N);
};

/**
 *  phi
 */
CImg<double> Daubechies::phi(int totLvl, int resLvl,int n){
  
  int nL = (totLvl-resLvl);
  CImgList<double> cPhi;
  
  cPhi.assign(nL+1);
  cPhi(0).assign(1<<resLvl,1,1,1, 0.);
  for (int k=1; k<=nL; k++) cPhi(k).assign(1<<(resLvl+k-1),1,1,1, 0.);
  cPhi(0)(n) = 1.;
  return compose1D(cPhi,nL,nL,1)*sqrt(1<<totLvl);
};

/**
 *  psi
 */
CImg<double> Daubechies::psi(int totLvl, int resLvl,int n){
  
  int nL = (totLvl-resLvl);
  CImgList<double> cPsi;
  
  cPsi.assign(nL+1);
  cPsi(0).assign(1<<resLvl,1,1,1, 0.);
  for (int k=1; k<=nL; k++) cPsi(k).assign(1<<(resLvl+k-1),1,1,1, 0.);
  cPsi(1)(n) = 1.;
  return compose1D(cPsi,nL,nL,1)*sqrt(1<<totLvl);
};

/**
 *  listWavelets
 */
void Daubechies::listWavelets(CImgList<double> &Phi, CImgList<double> &Psi, int nLevel){
  
  int dim = 1<<nLevel;
  CImgList<double> temp(nLevel+1);
  
  //--allocate temporary coeff pyramid
  temp(0).assign(dim>>nLevel,1,1,1,0.);
  for (int k=0; k<nLevel; k++) {
    temp(k+1).assign(dim>>(nLevel-k),1,1,1,0.);
  }
  
  //--allocate & compute details values
  Psi.assign(nLevel);
  for (int k=0; k<nLevel; k++) {
    //--at each scale
    Psi(k).assign(dim, temp(k+1).width());
    //--compute (periodized) translations
    for (int n=0; n<Psi(k).height(); n++) {
      //-reset coeffs
      for (int z=0; z<=nLevel; z++) temp(z).fill(0.);
      temp(k+1)(n) = 1.;
      //-compute function value through filter bank
      Psi(k).get_shared_line(n) = coeff2image1D(temp,nLevel);
    }
  }
  
  //--reset again temp
  for (int z=0; z<=nLevel; z++) temp(z).fill(0.);
  
  //--allocate & compute coarse values
  Phi.assign(nLevel);
  for (int k=0; k<nLevel; k++) {
    //--at each scale
    Phi(k).assign(dim, temp(0).width());
    //--compute (periodized) translations
    for (int n=0; n<Phi(k).height(); n++) {
      //-reset coarse coeffs
      temp(0).fill(0.);
      temp(0)(n) = 1.;
      //-compute function value through filter bank
      Phi(k).get_shared_line(n) = coeff2image1D(temp,nLevel-k);
    }
    //--modify temp to compute next scale
    temp.pop_front();
    temp(0).assign(temp(1).width(),1,1,1, 0.);
  }

};

/**
 *  niceView :D
 */
CImg<double> Daubechies::niceView(const CImg<double>& In, int level){
  
  CImgList<double> coeff = image2coeff2D(In, level);
  CImg<double> out(In.width(), In.height(),1,1, -1000), proj;
  int off;
  
  proj = projection2D(coeff,level);
  
  //-fill with approx
  cimg_forXY(proj,p,q){
    coeff(0).normalize(0,255);
    out(p,q) = coeff(0)(p,q);
  }
  
  //-then coeff
  for (unsigned int s=1; s<coeff.size(); s++) {
    off = coeff(s).width();
    coeff(s).normalize(0,255);
    //-add horizontal
    //coeff(s).get_shared_plane(0).normalize(0,255);
    cimg_forXY(coeff(s),p,q){
      out(p+off,q) = coeff(s)(q,p,0);
    }
    //-add vertical
    //coeff(s).get_shared_plane(1).normalize(0,255);
    cimg_forXY(coeff(s),p,q){
      out(p,q+off) = coeff(s)(q,p,1);
    }
    //-add diagonal
    //coeff(s).get_shared_plane(2).normalize(0,255);    
    cimg_forXY(coeff(s),p,q){
      out(p+off,q+off) = coeff(s)(q,p,2);
    }
  }
  return out;
};

/*--------------------------*/
/* Projection/Interpolation */
/*--------------------------*/

/**
 *  projection1D
 */
CImg<double> Daubechies::projection1D(const CImg<double>& In, int level){
  
  int n0 = (int)log2(In.width()), off;
  CImg<double> coarse, Phi, res;
  
  if (level>0) {      
    //-decompose and keep coarse
    coarse = image2coeff1D(In,level)(0);
    //-Phi
    Phi = phi(n0,n0-level,0);
    for (int k=0; k<level; k++) Phi = decimate(Phi);
    //-line
    res.assign(convolve(coarse,0,Phi,0,off,1));
  }
  else {
    res = In;
  }
  return res;
};

CImg<double> Daubechies::projection1D(const CImgList<double>& cIn, int level){
  
  int n0 = cIn.size()-1, off;
  CImg<double> coarse, Phi, res;
  
  //-recompose and keep coarse
  coarse = compose1D(cIn,off,n0-level,1);
  
  if(level>0){
    //-Phi
    Phi = phi(n0+(int)log2(cIn(0).width()),n0+(int)log2(cIn(0).width())-level,0);
    for (int k=0; k<level; k++) Phi = decimate(Phi);
    //-lines
    res.assign(convolve(coarse,0,Phi,0,off,1));
  }
  else {
    res = coarse*sqrt(coarse.width());
  }
  return res;  
};

/**
 *  projection2D
 */
CImg<double> Daubechies::projection2D(const CImg<double>& In, int level){
  
  int nW = (int)log2(In.width()), nH = (int)log2(In.height()), off;
  CImg<double> coarse, PhiW, PhiH, res;
  
  //-check
  if ((nW<level)||(nH<level)) {
    cout << "\t /!\\ Impossible projection requested (" << level <<"), corrected to " << min(nW,nH) << endl;
    level = min(nW,nH);
  }
  else if (level<0){
    cout << "\t /!\\ Impossible projection requested (" << level <<"), corrected to 0." << endl;
    level = 0;
  }
  
  if (level>0) {      
    //-decompose and keep coarse
    coarse = image2coeff2D(In,level)(0);
    //-Phi
    PhiW = phi(nW,nW-level,0);
    PhiH = phi(nH,nH-level,0);
    for (int k=0; k<level; k++){
      PhiW = decimate(PhiW);
      PhiH = decimate(PhiH);
    }
    //-lines
    res.assign(coarse.width(),coarse.height(),1,1, 0.);
    for (int p=0; p<res.height(); p++) {
      res.get_shared_line(p) = convolve(coarse.get_line(p),0,PhiW,0,off,1);
    }
    //-columns
    coarse = res.transpose();
    res.fill(0.);
    for (int p=0; p<res.height(); p++) {
      res.get_shared_line(p) = convolve(coarse.get_line(p),0,PhiH,0,off,1);
    }
    res.transpose();
  }
  else {
    res = In;
  }
  return res;  
};

CImg<double> Daubechies::projection2D(const CImgList<double>& cIn, int level){
  
  int n0 = cIn.size()-1, off;
  CImg<double> coarse, PhiW, PhiH, res;
  
  //-check
  if (level<0) {
    cout << "\t /!\\ Impossible projection requested (" << level <<"), corrected to " << 0 << endl;
    level = 0;
  }
  else if (n0<level){
    cout << "\t /!\\ Impossible projection requested (" << level <<"), corrected to " << n0 << endl;
    level = n0;
  }
  
  //-recompose and keep coarse
  coarse = compose2D(cIn,n0-level);
  
  if(level>0){
    //-Phi
    PhiW = phi(n0+(int)log2(cIn(0).width()),n0+(int)log2(cIn(0).width())-level,0);
    PhiH = phi(n0+(int)log2(cIn(0).height()),n0+(int)log2(cIn(0).height())-level,0);
    for (int k=0; k<level; k++){ 
      PhiW = decimate(PhiW);
      PhiH = decimate(PhiH);
    }
    //-lines
    res.assign(coarse.width(),coarse.height(),1,1, 0.);
    for (int p=0; p<res.height(); p++) {
      res.get_shared_line(p) = convolve(coarse.get_line(p),0,PhiW,0,off,1);
    }
    //-columns
    coarse.assign(res.transpose());
    res.fill(0.);
    for (int p=0; p<res.height(); p++) {
      res.get_shared_line(p) = convolve(coarse.get_line(p),0,PhiH,0,off,1);
    }
    res.transpose();
  }
  else {
    res = coarse*sqrt(coarse.size());
  }
  return res;  
};

/**
 *  interp1D
 */
CImg<double> Daubechies::interp1D(const CImg<double>& In){

  int off;
  
  CImg<double> out = expand(In);
  CImg<double> temp = convolve(In, 0, hi, hiOff, off, 1);
  for (off=0; off<In.width(); off++) out(2*off+1) = temp(off);
  temp.assign();
  return out;
};

/**
 *  interp2D
 */
CImg<double> Daubechies::interp2D(const CImg<double>& In){
  
  CImg<double> out(2*In.width(),2*In.height(),1,1,0.);
  
  //--treat rows
  for (int r=0; r<In.height(); r++) out.get_shared_line(2*r) = interp1D(In.get_line(r));
  out.transpose();
  //--then columns
  for (int r=0; r<out.height(); r++) out.get_shared_line(r) = interp1D(decimate(out.get_line(r)));
  
  return out.transpose();
};

/**
 *  diagInterp2D
 */
CImg<double> Daubechies::diagInterp2D(const CImg<double>& In){
  
  CImg<double> out(In.width(),In.height(),1,1,0.);
  
  int off;
  
  //--treat rows
  for (int r=0; r<In.height(); r++) out.get_shared_line(r) = convolve(In.get_line(r), 0, hi, hiOff, off, 1);
  out.transpose();
  //--then columns
  for (int r=0; r<out.height(); r++) out.get_shared_line(r) = convolve(out.get_line(r), 0, hi, hiOff, off, 1);
  
  return out.transpose();
};

/**
 *  diagTranspInterp2D
 */
CImg<double> Daubechies::diagTranspInterp2D(const CImg<double>& In){
  
  CImg<double> out(In.width(),In.height(),1,1,0.);
  
  int off;
  
  //--treat rows
  for (int r=0; r<In.height(); r++) out.get_shared_line(r) = convolve(In.get_line(r), 0, hib, hibOff, off, 1);
  out.transpose();
  //--then columns
  for (int r=0; r<out.height(); r++) out.get_shared_line(r) = convolve(out.get_line(r), 0, hib, hibOff, off, 1);
  
  return out.transpose();
};




/*==============================*/
/* COEFF DECAY / MODULUS MAXIMA */
/*==============================*/

//-------!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//   need to be modified : rectangular dimensions, ***transposed details***
//-------!!!!!!!!!!!!!!!!!!!!!!!!!!!!

CImgList<double> Daubechies::get_Leaders1D(const CImgList<double>& wavCoeffs, CImgList<double>& vanishMap, int startLevel){
  
  int inputLevel = wavCoeffs.size();
  int resLevel;
  CImgList<double> result, absCoeffs;
  
  //-there will as many dominating coeffs as the total numer of coeffs at the different scales
  //-we also precompute abs(coeffs) at the explored scales
  resLevel = inputLevel-startLevel;
  result = CImgList<double>(resLevel);
  absCoeffs = CImgList<double>(resLevel);
  vanishMap = CImgList<double>(resLevel);
  for (int k=0; k<resLevel; k++){
    absCoeffs(k) = wavCoeffs(startLevel+k).get_abs();
    result(k) = CImg<double>(wavCoeffs(startLevel+k).width(), 1, 1, 1, 0.);
    vanishMap(k) = CImg<double>(wavCoeffs(startLevel+k).width(), 1, 1, 1, 1.);
  }
  //-exploring each scale including smallest
  for (int k=0; k<resLevel; k++) {
    cout << "+++ scale "<<k+startLevel<<endl;
    for (int q=0; q<result(k).width(); q++) {
      //cout << q << endl;
      (result(k))(q) = findLeaders1D(absCoeffs, vanishMap, k, q); 
    }
  }
  return result;
};

CImgList<double> Daubechies::get_Leaders2D(const CImgList<double>& wavCoeffs, CImgList<double>& vanishMap, int startLevel){
  
  int inputLevel = wavCoeffs.size();
  int resLevel;
  CImgList<double> result, absCoeffs;
  
  //-there will as many dominating coeffs as the total numer of coeffs at the different scales
  //-we also precompute abs(coeffs) at the explored scales
  resLevel = inputLevel-startLevel;
  result = CImgList<double>(resLevel);
  absCoeffs = CImgList<double>(resLevel);
  vanishMap = CImgList<double>(resLevel);
  for (int k=0; k<resLevel; k++){
    absCoeffs(k).assign(wavCoeffs(startLevel+k).get_abs());
    result(k) = CImg<double>(wavCoeffs(startLevel+k).width(), wavCoeffs(startLevel+k).height(), 3, 1, 0.);
    vanishMap(k) = CImg<double>(wavCoeffs(startLevel+k).width(), wavCoeffs(startLevel+k).height(), 3, 1, 1.);
  }
  //-exploring each scale even smallest
  cout << "++++>> scales "<< flush;
  for (int k=0; k<resLevel; k++) {
    cout << k+startLevel << " / " <<flush;
    for (int p=0; p<result(k).width(); p++) {
      for (int q=0; q<result(k).height(); q++) {
        (result(k))(p,q,0) = findLeaders2D(absCoeffs, vanishMap, k, p, q, 0);
        (result(k))(p,q,1) = findLeaders2D(absCoeffs, vanishMap, k, p, q, 1);
        (result(k))(p,q,2) = findLeaders2D(absCoeffs, vanishMap, k, p, q, 2);
      }
    }
  }
  cout << endl;
  return result;
}

double Daubechies::findLeaders1D(const CImgList<double>& Coeffs, CImgList<double>& vanishMap, int level, int q){
  
  double result = Coeffs(level)(q), below, tol = 1e-15;

  //-si pas un bord
  if ((q>0)&&(q<Coeffs(level).width()-1)) {
    if(level<Coeffs.size()-1){
      //-check below middle
      below = lookBelow1D(Coeffs, level, q);
      if (result < below) result = below;      
      //-/!\---update vanish map
      if (result < tol) vanishMap(level)(q) = 0.;
    }
    else {
      if (result < tol) vanishMap(level)(q) = 0.;
    }
    //check left
    if (result<Coeffs(level)(q-1)) result = Coeffs(level)(q-1);
    //check right
    if (result<Coeffs(level)(q+1)) result = Coeffs(level)(q+1);
    //-check below if possible
    if(level<Coeffs.size()-1){
      //-below left
      below = lookBelow1D(Coeffs, level, q-1);
      if (result < below) result = below;
      //-and right
      below = lookBelow1D(Coeffs, level, q+1);
      if (result < below) result = below;
    }
  }
  else vanishMap(level)(q) = 0.;
  
  return result;
};

double Daubechies::findLeaders2D(const CImgList<double>& Coeffs, CImgList<double>& vanishMap, int level, int p, int q, int z){
  
  double result = Coeffs(level)(p,q,z), below;
  double tol = 1e-15;
  
  //-si pas un bord
  if ((p>0)&&(p<Coeffs(level).width()-1)&&(q>0)&&(q<Coeffs(level).height()-1)) {
    //-deal with middle & below middle first
    if(level<Coeffs.size()-1){
      //-check below middle
      below = lookBelow2D(Coeffs, level, p, q, z);
      if (result < below) result = below;
      //-/!\---update vanish map
      if (result < tol) vanishMap(level)(p, q, z) = 0.;
    }
    else {
      if (result < tol) vanishMap(level)(p, q, z) = 0.;
    }
    //check left
    if (result < Coeffs(level)(p-1,q,z)) result = Coeffs(level)(p-1,q,z);
    //check right
    if (result < Coeffs(level)(p+1,q,z)) result = Coeffs(level)(p+1,q,z);
    //check up
    if (result < Coeffs(level)(p,q-1,z)) result = Coeffs(level)(p,q-1,z);
    //check down
    if (result < Coeffs(level)(p,q+1,z)) result = Coeffs(level)(p,q+1,z);
    //check left up
    if (result < Coeffs(level)(p-1,q-1,z)) result = Coeffs(level)(p-1,q-1,z);
    //check left right 
    if (result < Coeffs(level)(p+1,q-1,z)) result = Coeffs(level)(p+1,q-1,z);
    //check left down
    if (result < Coeffs(level)(p-1,q+1,z)) result = Coeffs(level)(p-1,q+1,z);
    //check right down
    if (result < Coeffs(level)(p+1,q+1,z)) result = Coeffs(level)(p+1,q+1,z);
    //-check below neighbors if possible
    if(level<Coeffs.size()-1){
      //-below left
      below = lookBelow2D(Coeffs, level, p-1, q, z);
      if (result < below) result = below;
      //-below right
      below = lookBelow2D(Coeffs, level, p+1, q, z);
      if (result < below) result = below;
      //-below up
      below = lookBelow2D(Coeffs, level, p, q-1, z);
      if (result < below) result = below;
      //-and down
      below = lookBelow2D(Coeffs, level, p, q+1, z);
      if (result < below) result = below;
      //-below left up
      below = lookBelow2D(Coeffs, level, p-1, q-1, z);
      if (result < below) result = below;
      //-below right up
      below = lookBelow2D(Coeffs, level, p+1, q-1, z);
      if (result < below) result = below;
      //-below left down
      below = lookBelow2D(Coeffs, level, p-1, q+1, z);
      if (result < below) result = below;
      //-and right down
      below = lookBelow2D(Coeffs, level, p+1, q+1, z);
      if (result < below) result = below;
    }
  }
  else vanishMap(level)(p, q, z) = 0.;
  return result;
};

double Daubechies::lookBelow1D(const CImgList<double>& Coeffs, int level, int q){
  
  double left, right;
  double rl, rr;
  int currentLevel = level+1;
  int il = q<<1;
  int ir = (q<<1)+1;
  
  //-si on est en bas
  if (currentLevel==Coeffs.size()-1) {
    //cout << "down !!"<<endl;
    left = Coeffs(currentLevel)(il);
    right = Coeffs(currentLevel)(ir);
  }
  //-sinon
  else {
    rl = lookBelow1D(Coeffs, currentLevel, il);
    left = (Coeffs(currentLevel)(il)>rl)?Coeffs(currentLevel)(il):rl;
    rr = lookBelow1D(Coeffs, currentLevel, ir);
    right = (Coeffs(currentLevel)(ir)>rr)?Coeffs(currentLevel)(ir):rr;
  }
  return (left>right)?left:right;
};

double Daubechies::lookBelow2D(const CImgList<double>& Coeffs, int level, int p, int q, int z){
  
  double upLeft, upRight, downLeft, downRight, up, down;
  double rul, rur, rdl, rdr;
  int currentLevel = level+1;
  int il = p<<1, ir = (p<<1)+1, iu = q<<1, id = (q<<1)+1;
  
  //-si on est en bas
  if (currentLevel==Coeffs.size()-1) {
    //if (z==0) cout << "down !! "<< il << " ; " << iu <<endl;
    upLeft = Coeffs(currentLevel)(il,iu,z);
    //if (z==0) cout << "down !! "<< ir << " ; " << iu <<endl;
    upRight = Coeffs(currentLevel)(ir,iu,z);
    //if (z==0) cout << "down !! "<< il << " ; " << id <<endl;
    downLeft = Coeffs(currentLevel)(il,id,z);
    //if (z==0) cout << "down !! "<< ir << " ; " << id <<endl;
    downRight = Coeffs(currentLevel)(ir,id,z);
  }
  //-sinon
  else {
    //if (z==0) cout << "---[lu]" << il << " ; " << iu << endl; 
    rul = lookBelow2D(Coeffs, currentLevel, il,iu,z);
    upLeft = (Coeffs(currentLevel)(il,iu,z)>rul)?Coeffs(currentLevel)(il,iu,z):rul;
    //if (z==0) cout << "---[ru]" << ir << " ; " << iu << endl; 
    rur = lookBelow2D(Coeffs, currentLevel, ir,iu,z);
    upRight = (Coeffs(currentLevel)(ir,iu,z)>rur)?Coeffs(currentLevel)(ir,iu,z):rur;
    //if (z==0) cout << "---[ld]" << il << " ; " << id << endl; 
    rdl = lookBelow2D(Coeffs, currentLevel, il,id,z);
    downLeft = (Coeffs(currentLevel)(il,id,z)>rdl)?Coeffs(currentLevel)(il,id,z):rdl;
    //if (z==0) cout << "---[rd]" << ir << " ; " << id << endl; 
    rdr = lookBelow2D(Coeffs, currentLevel, ir,id,z);
    downRight = (Coeffs(currentLevel)(ir,id,z)>rdr)?Coeffs(currentLevel)(ir,id,z):rdr;
  }
  
  up = (upLeft>upRight)?upLeft:upRight;
  down = (downLeft>downRight)?downLeft:downRight;
  
  return (up>down)?up:down;
};

double** Daubechies::get_StructFunc1D(const CImgList<double>& Coeffs, const CImgList<double>& vanishMap, int order, int nVal){
  
  CImgList<double> tempCoeffs = Coeffs;
  double ** result;
  double step=2.*(double)order/(double)(nVal-1);
  CImg<double> temp;
  
  result = new double*[nVal];
  for (int k=0; k<nVal; k++) result[k] = new double[tempCoeffs.size()];
  
  //---for each order
  for (int k=0; k<nVal; k++) {
    //---at each scale
    for (int p=0; p<tempCoeffs.size(); p++) {
      result[k][p] = 0.;
      //---compute temp
      temp.assign(tempCoeffs(p).get_pow(k*step-(double)order));
      cimg_forX(temp, x){
        //---store value
        if (vanishMap(p)(x)) result[k][p] += temp(x);
      }
      result[k][p] /= temp.size(); 
    }
  }
  return result;
};

double** Daubechies::get_StructFunc2D(const CImgList<double>& Coeffs, const CImgList<double>& vanishMap, int order, int nVal, int type){
  
  CImgList<double> tempCoeffs = Coeffs;
  double ** result;
  double step=2.*(double)order/(double)(nVal-1);
  CImg<double> temp;
  
  //---allocate result
  result = new double*[nVal];
  for (int k=0; k<nVal; k++) result[k] = new double[tempCoeffs.size()];
  
  //---for each order
  for (int k=0; k<nVal; k++) {
    //---at each scale
    for (int q=0; q<tempCoeffs.size(); q++) {
      result[k][q] = 0.;
      //---compute temp
      temp.assign(tempCoeffs(q).get_slice(type).pow(k*step-(double)order));
      cimg_forXY(temp,x,y){
        if (vanishMap(q)(x,y,type)) result[k][q] += temp(x,y);
      }
      result[k][q] /= temp.size(); 
    }
  }
  return result;
};

/*===============*/
/* POWER SPECTRA */
/*===============*/
CImg<double> Daubechies::iso2DWavSpectrum(const CImgList<double> &coeffs){

  //----all we gotta do is
  //+ for each scale, sum square of every coeff
  //+ multiply by k=2^scale
  //----------------------  
  CImg<double> res(coeffs.size(),2,1,1, 0.);
  
  for (int s=1; s<coeffs.size(); s++) {
    res(s,0) = (double)(1<<(s-1));
    res(s,1) = .5*coeffs(s).get_sqr().sum();
  }
  return res;  
};

/*===============*/
/* VARIOUS TOOLS */
/*===============*/

CImg<double> Daubechies::expand(const CImg<double>& A){
 
  CImg<double> res(2*A.width(),1,1,1,0);
 
  for (int k=0; k<A.width(); k++) res(2*k) = A(k);
  return res;
};

CImg<double> Daubechies::decimate(const CImg<double>& A){

  int Rdim = A.width()>>1;
  CImg<double> res;
  
  //----CUIDAOOOO ! primary version, only finite signal s[n], n>=0
  Rdim += (A.width()%2==0)?0:1;
  res = CImg<double>(Rdim,1,1,1,0);
  for (int k=0; k<Rdim; k++) res(k) = A(2*k);
  return res;
};
    
CImg<double> Daubechies::convolve(const CImg<double>& signal, const int& Soffset,
                                  const CImg<double>& filtre, const int& Foffset,
                                  int& Roffset, int border){

  int Rdim;
  CImg<double> result;
  
  if (border){
    //-initialisation
    Roffset = Soffset;
    Rdim = signal.width();
    result = CImg<double>(Rdim,1,1,1,0);
   
    //-remplissage
    for (int n=Roffset; n<Rdim+Roffset; n++) {
      for (int k=Foffset; k<filtre.width()+Foffset; k++) {
        //result(n-Roffset) += ( (n-k<Soffset)||(n-k>=Soffset+signal.width()) )?signal(period(n-k-Soffset,signal.width()))*filtre(k-Foffset):signal(n-k-Soffset)*filtre(k-Foffset);
        result(n-Roffset) += signal(period(n-k-Soffset,signal.width()))*filtre(k-Foffset);
        //cout<<"["<<n<<"] n-k: "<<n-k<<" k:"<<k<<"\t n-k-Soffset: "<<n-k-Soffset<<" period: "<<period(n-k-Soffset,signal.width())<<endl;
      }
    }
  } else {
    //-initialisation
    Roffset = Soffset + Foffset;
    Rdim = signal.width() + filtre.width();
    result = CImg<double>(Rdim,1,1,1,0);
    //-remplissage
    for (int n=Roffset; n<Rdim+Roffset; n++) {
      for (int k=Foffset; k<filtre.width()+Foffset; k++) {
        result(n-Roffset) += ( (n-k<Soffset)||(n-k>=Soffset+signal.width()) )?0:signal(n-k-Soffset)*filtre(k-Foffset);
      }
    }
  }
  return result;
};

CImg<double> Daubechies::add(const CImg<double>& a, const int& Aoffset, 
                             const CImg<double>& b, const int& Boffset, 
                             int& Roffset){
  
  CImg<double> res;
  int aMax = a.width()+Aoffset-1, bMax = b.width()+Boffset-1;
  int rMax = (aMax>bMax)?aMax:bMax;
    
  Roffset = (Aoffset<Boffset)?Aoffset:Boffset;
  res = CImg<double>(rMax-Roffset+1,1,1,1,0);
  for (int k=0; k<a.width(); k++) res(k-Roffset+Aoffset) += a(k);
  for (int k=0; k<b.width(); k++) res(k-Roffset+Boffset) += b(k);  
  return res;
};

inline int Daubechies::period(int ind, int max){
  return (ind<0)?((max + ind%max)%max):ind%max;
};
