/*
 *  OpFlowTools.h
 *  
 **
 */

#ifndef DEF_OFTOOLS
#define DEF_OFTOOLS
 

/*-- file (image, fields) input/output --*/
#include "OpFlowTools_io.h"
/*-- basic field operation (gradients ...) --*/
#include "OpFlowTools_fieldOp.h"
/*-- structure function computation --*/
#include "OpFlowTools_structFun.h"
/*-- visualization functions --*/
#include "OpFlowTools_visu.h"

cimg_library::CImg<double>  butterworth(const cimg_library::CImg<double> &Im,double k0,double ratioSpec, int order);
cimg_library::CImg<double> productInFourier(const cimg_library::CImg<double> &im1,const cimg_library::CImg<double> &im2);

/*!
 @abstract  Low pass filtering in Fourier domain.
 @param      In inoput image
 @param      cutOff cuttOff frequency (signal spatialy normalized in [0,1]).
 @param      factor subsampling factor.
 @result     The low passed image.
 */
cimg_library::CImg<double> LowPassDoor(const cimg_library::CImg<double> &In,double cutOff,int factor);
cimg_library::CImg<double> LowPassDoor(const cimg_library::CImg<double> &In,double cutOff);
cimg_library::CImg<double> LowPassDoorWithAddBandLimitedSignal(const cimg_library::CImg<double> &In,const cimg_library::CImg<double> &Im2Add);
/*!
 @abstract   Interpolation by sinus cardinal at resolution 2^factor times higher.
 @param      In inoput image
 @param      factor resolution factor.
 @result     The interpolated image.
 */
cimg_library::CImg<double> interpSinc(const cimg_library::CImg<double> &In,int factor);

/*!
    @abstract   Use vorticity of the given motion field to color given image.
    @param      uv A motion field from which vorticity is computed 
    @param      background The image to be colored
    @param      mask An option mask. Will be renormalized to (0,1), so any image will do it.
    @param      curlbound An optional positive threshold for vorticity (i.e. keep values |curl|<=curlbound).
    @param      darkshift An optional positive value to shift values of background image towards white (i.e. ligthen dark regions). Must be < 1.
    @param      sigma An optional smoothing parameter for (gradient) vorticity computation. 
    @result     The colored image.
*/

cimg_library::CImg<double> curlPaint(const cimg_library::CImg<double> &uv,
                                     const cimg_library::CImg<double> &background,
                                     const cimg_library::CImg<double> &mask=cimg_library::CImg<double>::empty(),
                                     double curlbound=0.,
                                     double darkshift=0.125,
                                     double sigma=.5);

cimg_library::CImg<double> curlImage(const cimg_library::CImg<double> &uv,
                                     const cimg_library::CImg<double> &mask=cimg_library::CImg<double>::empty(),
                                     double curlbound=0.,
                                     double sigma=.5);

/*!
 @abstract   Use vector-norm of the given motion field to color given image.
 @param      uv A motion field from which norm is computed 
 @param      background The image to be colored
 @param      mask An option mask. Will be renormalized to (0,1), so any image will do it.
 @param      normbound An optional positive threshold for norm (i.e. keep values norm<=normbound).
 @param      darkshift An optional positive value to shift values of background image towards white (i.e. ligthen dark regions). Must be < 1.
 @result     The colored image.
 */

cimg_library::CImg<double> normPaint(const cimg_library::CImg<double> &uv,
                                     const cimg_library::CImg<double> &background,
                                     const cimg_library::CImg<double> &mask=cimg_library::CImg<double>::empty(),
                                     double normbound=0.,
                                     double darkshift=0.125);

cimg_library::CImg<double> normImage(const cimg_library::CImg<double> &uv,
                                     const cimg_library::CImg<double> &mask=cimg_library::CImg<double>::empty(),
                                     double normbound=0.);

void  FourierLerayProjector(cimg_library::CImgList<double> &fIm_u, cimg_library::CImgList<double> &fIm_v);
void LerayProjector( cimg_library::CImg<double> &uv);
void fourierBrownianSampling(cimg_library::CImg<double> &uv, double zeta, double minS, double maxS, bool divZero);
/*!
 @abstract   Caluculates the fractional derivative/integration of a given motion field.
 @param      Im A motion field  
 @param      zeta The exponent defining the derivative/integration in Fourier space
 @param      LerayProj Leray projection
 @result     The motion field fractional derivative/integration.
 */
cimg_library::CImg<double>  fracDerivOrInt_1Band(const cimg_library::CImg<double> &Im,double zeta);
cimg_library::CImg<double>  fracDerivOrInt(const cimg_library::CImg<double> &Im,double zeta);
cimg_library::CImg<double>  fracDerivOrInt(const cimg_library::CImg<double> &Im,double zeta, bool LerayProj);

/*!
 @abstract   Caluculates the fractional integration/derivative of a given motion field.
 @param      Im A motion field  
 @param      zeta Minus the exponent defining the derivative/integration in Fourier space
 @param      LerayProj Leray projection
 @result     The motion field fractional integration/derivative.
 */
cimg_library::CImg<double>  fracIntOrDeriv(const cimg_library::CImg<double> &Im,double zeta);
cimg_library::CImg<double>  fracIntOrDeriv(const cimg_library::CImg<double> &Im,double zeta, bool LerayProj);

/*!
 @abstract   Caluculates the curl and fractional derivative of a scalar field.
 @param      Im A scalar field  
 @param      zeta Exponent defining the derivative/integration in Fourier space
  @result    Curl of the scalar field fractional derivative/integration.
 */
cimg_library::CImg<double>  fracDerivOrInt_Curl(const cimg_library::CImg<double> &Im,double zeta);

/*!
 @abstract   Caluculates the curl and fractional derivative of a scalar field + its grad .
 @param      Im A scalar field
 @param      zeta Exponent defining the derivative/integration in Fourier space
 @result    Curl of the scalar field fractional derivative/integration and gardient (4 images).
 */
cimg_library::CImg<double>  fracDerivOrInt_CurlAndGrad(const cimg_library::CImg<double> &Im,double zeta);
/*!
 @abstract   Caluculates grad .
 @param      Im A scalar field
 @result     gradient (2 images).
 */
cimg_library::CImg<double>  gradFourier(const cimg_library::CImg<double> &Im);
/*!
 @abstract   Caluculates the curl adjoint and fractional derivative of a vectorial field.
 @param      Im A vectorial field  
 @param      zeta Exponent defining the derivative/integration in Fourier space
 @param      twoParts Two parts of curl adjoint returned separately
 @result     Curl adjoint of the vectorial field fractional derivative/integration.
 */
cimg_library::CImg<double>  fracDerivOrInt_CurlAdjoint(const cimg_library::CImg<double> &Im,double zeta);
cimg_library::CImg<double>  fracDerivOrInt_CurlAdjoint(const cimg_library::CImg<double> &Im,double zeta,bool twoParts);

/*!
 @abstract   Caluculates in 2 parts (for normalized version) the adjoint of curl dual and fractional derivative of a vectorial field.
 @param      Im A vectorial field  
 @param      zeta Exponent defining the derivative/integration in Fourier space
 @result     Curl dual of the vectorial field fractional derivative/integration.
 */
cimg_library::CImg<double>  fracDerivOrInt_CurlDualAdjoint2Part(const cimg_library::CImg<double> &Im,double zeta);

/*!
 @abstract   Normalize Fourier-based div-free wavelets Psi \times Psi.
 @param      in A vectorial field  
 @result     Normalized field.
 */
cimg_library::CImg<double> normalizeFourierDivFreeAnisWav(const cimg_library::CImg<double> &in);

/*!
 @abstract   Normalize Fourier-based div-free wavelets Psi \times Psi and Psi \times Phi.
 @param      in A vectorial field  
 @result     Normalized field.
 */
cimg_library::CImg<double> normalizeFourierDivFreeAnisWavV0(const cimg_library::CImg<double> &in);

/*!
 @abstract   Normalize Fourier-based div-free dual  wavelets Psi \times Psi.
 @param      in A vectorial field  
 @result     Normalized field.
 */
cimg_library::CImg<double> normalizeFourierDivFreeAnisWavDual(const cimg_library::CImg<double> &in);

/*!
 @abstract   Accelerate matrix multiplication (Clapack)
 */
//cimg_library::CImg<double> operator*(const cimg_library::CImg<double>&A, const cimg_library::CImg<double> &B);

 inline int period(int ind, int max);


#endif