/*
 *  Anisotropic Fractional Divergence-free aniWavOpticalFlow.cpp
 *
 *
 *  Written by Patrick Héas.
 *  Copyright 2014 INRIA. All rights reserved.
 *
 */

#include <iomanip>
#include <sstream>
#include "aniWavOpticalFlow.h"
#include <time.h>
#include <random>

using namespace cimg_library;
using namespace std;


/********************
 *
 *  WavOpticalFlow class
 *  Fractional and DivFree options
 *
 *******************/


//=================
//------constructor
//=================

WavOpticalFlow::WavOpticalFlow(CImg<double> Im0,
                               CImg<double> Im1,
                               CImg<double> ImMask,
                               CImg<double> ImUVreal,
                               int nD,
                               int nE,
                               int nI,
                               int nM,
                               int wT,
                               double sig,
                               int reg,
							   double g,
							   double z,
							   double zIn,
							   bool frac,
							   bool div0,
							   bool per,
							   double minPwr,
							   double maxPwr,
							   double c_stdev,
							   double t,
                               bool app,
                               string coeffDir_) :
I0(Im0), I1(Im1), UV(Im0.width(),Im0.height(),2,1,0.), Mask(ImMask),UVreal(ImUVreal),
dimW(Im0.width()), dimH(Im0.height()), nDecomp(nD), nEstim(nE), nInit(nI), currentLvl(0), wav(nM,wT),
regul(reg), sigma(sig), approx(app), coeffDir(coeffDir_)
{
    
    //-allocation wavelet coeff /!\ transposed dimensions !!
    cU.assign(I0.height(),I0.width(),1,1, 0.);
    if(frac){cV.assign(I0.height(),I0.width(),1,1, 0.);}
    //-then we can also check & correct nEstim value
    if (nEstim>nDecomp) {
        cout << "\t /!\\ Impossible estimation requested (" << nEstim <<"), corrected to " << nDecomp << "(full resolution estimation)"<< endl;
        nEstim = nDecomp;
    }
    else if (nEstim<0){
        cout << "\t /!\\ Impossible estimation requested (" << nEstim <<" levels), corrected to " << 0 << "(coarse only)"<< endl;
        nEstim = 0;
    }
    
    //-check mask, remove if not necessary.
    if ((Mask.min()==1.)&&(Mask.max()==1.)) {Mask.assign();}
    
    //-regularization weights
    gamma = g;
    zeta=z;
    zetaIn=zIn;
    //-original scale
    n0w = (int)log2(dimW);
    n0h = (int)log2(dimH);
    n0 = min(n0w,n0h);
    
    //-total number of coeff
    if(frac){
        nTotCoeff = 1<<(n0w+n0h+2*(nEstim-nDecomp));
        //-initialize theta
        theta.assign(1<<(n0h-nDecomp+nEstim),1<<(n0w-nDecomp+nEstim),2,1, 0.);
        thetaReal.assign(1<<(n0h-nDecomp+nEstim),1<<(n0w-nDecomp+nEstim),2,1, 0.);
        
    }
    else{
        nTotCoeff = 1<<(n0w+n0h+2*(nEstim-nDecomp));
        //-initialize theta
        theta.assign(1<<(n0h-nDecomp+nEstim),1<<(n0w-nDecomp+nEstim),1,1, 0.);
        thetaReal.assign(1<<(n0h-nDecomp+nEstim),1<<(n0w-nDecomp+nEstim),1,1, 0.);
    }
    //other
    divFree=div0;
    fraction=frac ;
    periodic=per;
    MinPwrRange=minPwr;
    MaxPwrRange=maxPwr;
    c=c_stdev;
    tau=t;
    if (divFree&&fraction){LerayProj=true;}else{LerayProj=false;}
};

//=================
//-deform-generatUV
//=================
CImg<double>  WavOpticalFlow::deformImByUVFracDivFree(){
	return deformImByUVFracDivFree(0, zeta,c,0.,0.);
}
CImg<double>  WavOpticalFlow::deformImByUVFracDivFree( double zetaGen){
	return deformImByUVFracDivFree(0,zetaGen,c,0.,0.);
}
CImg<double>  WavOpticalFlow:: deformImByUVFracDivFree(double zetaGen, double cc){
	return deformImByUVFracDivFree(0,zetaGen,cc,0.,0.);
    
}
CImg<double>  WavOpticalFlow:: deformImByUVFracDivFree(bool fourier, double zetaGen, double cc, double minS, double maxS){
    return deformAndReturnImByUVFracDivFree(fourier,zetaGen,cc,minS,maxS,0.)[0];
}
CImgList<double>  WavOpticalFlow:: deformAndReturnImByUVFracDivFree(bool fourier, double zetaGen, double cc, double minS, double maxS,double sigNoise){
    
    //generate fractional gaussian motion
    CImg<double> tmpUV;
    if (!fourier){
        //fourier-wavelet method
        if (maxS==0){tmpUV=(generatorUVFracDivFree(zetaGen,cc));}
        else{tmpUV=(generatorUVFracDivFree(zetaGen,cc,minS,maxS));}
    }
    else{
        //fourier method
        tmpUV.assign(I1.width(),I1.height(),2,1);
        fourierBrownianSampling(tmpUV,zetaGen,minS,maxS,divFree);
        tmpUV*=cc;
    }
    UVreal=tmpUV;
    
    //compensated image
    Recal_spline(I0,I1,tmpUV,3);
    
    CImgList<double> res(tmpUV,I0,I1,true);
    
    return res;
};
CImg<double>  WavOpticalFlow::generatorUVFracDivFree(){
	return generatorUVFracDivFree(zeta,c);
}
CImg<double>  WavOpticalFlow::generatorUVFracDivFree( double zetaGen){
	return generatorUVFracDivFree(zeta,c);
}
CImg<double>  WavOpticalFlow::generatorUVFracDivFree(double zetaGen, double cc){
    
	CImg<double> UVSample;UVSample.assign(UV,false);
	CImg<double> cUV;
	bool Leray=false;
    
//	srand((unsigned)time(NULL));
//	//srand(0);
//	long int *auxNoise2=new long int[1];
//	float gasdev(long *idum);
//	auxNoise2[0]=-(int)(((float)rand())/RAND_MAX*10000);
//	cUV.assign(cU.width(),cU.height(),2);cUV.fill(0.);
//	cimg_forXYZ(cUV,p,q,r){	cUV(p,q,r) = gasdev(auxNoise2);}
    cUV.noise(1.);
    
	//----------Rebuild UV
	//--------------------
    cUV(0,0,0)=0;cUV(0,0,1)=0;
    UVSample.get_shared_plane(0)=wav.anisCompose2D(cUV.get_shared_plane(0),nDecomp)*sqrt(dimW*dimH);
    UVSample.get_shared_plane(1)=wav.anisCompose2D(cUV.get_shared_plane(1),nDecomp)*sqrt(dimW*dimH);
    if (divFree){Leray=true;}
    UVSample=cc*fracDerivOrInt(UVSample,zetaGen,Leray)/(dimW*dimH);
 	cimg_forXYZ(thetaReal,p,q,r){	 thetaReal(p,q,r)=cUV(p,q,r)/(dimW*dimH)*cc;}
	int err = 0;
	double stdevSpec=0.;
	double tmpW=UVSample.width();
	write_Spectrum("toBeErased_SynthesisFBM_UVspectrum.txt",isotropicPSD(UVSample,stdevSpec,2.*M_PI/tmpW,0.1).width(),isotropicPSD(UVSample,stdevSpec,2.*M_PI/tmpW,0.1).get_shared_plane(0).data(),err);
	return UVSample;
};
CImg<double>  WavOpticalFlow::generatorUVFracDivFree(double zetaGen, double cc, double minS, double maxS){
    
	CImg<double> UVSample;UVSample.assign(UV,false);
	CImg<double> cUV;
	bool Leray=false;
//	srand((unsigned)time(NULL));
//	//srand(0);
//	long int *auxNoise2=new long int[1];
//	float gasdev(long *idum);
//	auxNoise2[0]=-(int)(((float)rand())/RAND_MAX*10000);
//	cUV.assign(cU.width(),cU.height(),2);cUV.fill(0.);
//	cimg_forXYZ(cUV,p,q,r){	cUV(p,q,r) = gasdev(auxNoise2);	}
//
    cUV.noise(1.);
	double scaleNorm,boundx,boundy;
	CImg<double> mask(dimW,dimH,1,1);
	mask.fill(0.);
    for (int sx=0;sx<log2(mask.width());sx++){
        for (int sy=0;sy<log2(mask.height());sy++){
            scaleNorm=sqrt(pow(2.,2*(n0w-sx-1))+pow(2.,2*(n0h-sy-1)));
            if (scaleNorm <= maxS && scaleNorm >= minS){
                if (sx==0){boundx=0;}else{boundx=(1<<sx);}
                if (sy==0){boundy=0;}else{boundy=(1<<sy);}
                for (int px=boundx;px<(1<<(sx+1));px++){
                    for (int py=boundy;py<(1<<(sy+1));py++){
                        mask(px,py) = 1;
                    }
                }
            }
        }
	}
	cUV.get_shared_plane(0).mul(mask);
	cUV.get_shared_plane(1).mul(mask);
	//----------Rebuild UV
	//--------------------
    cUV(0,0,0)=0;cUV(0,0,1)=0;
    UVSample.get_shared_plane(0)=wav.anisCompose2D(cUV.get_shared_plane(0),nDecomp)*sqrt(dimW*dimH);
    UVSample.get_shared_plane(1)=wav.anisCompose2D(cUV.get_shared_plane(1),nDecomp)*sqrt(dimW*dimH);
    if (divFree){Leray=true;}
    UVSample=cc*fracDerivOrInt(UVSample,zetaGen,Leray)/(dimW*dimH);
	int err = 0;
 	cimg_forXYZ(thetaReal,p,q,r){	 thetaReal(p,q,r)=cUV(p,q,r)/(dimW*dimH)*cc;}
	double stdevSpec=0.;
	double tmpW=UVSample.width();
	write_Spectrum("toBeErased_SynthesisFBM_UVspectrum.txt",isotropicPSD(UVSample,stdevSpec,2.*M_PI/tmpW,0.1).width(),isotropicPSD(UVSample,stdevSpec,2.*M_PI/tmpW,0.1).get_shared_plane(0).data(),err);
	return UVSample;
};


//=================
//-----------initUV
//=================
void WavOpticalFlow::initUVFracDivFree(const CImg<double> &UV0){
    
    int err = 0;
    double stdevSpec=0.;
    double tmpW=UV0.width();
    write_Spectrum("toBeErased_spectrumInit.txt",isotropicPSD(UV0,stdevSpec,2.*M_PI/tmpW,0.1).width(),isotropicPSD(UV0,stdevSpec,2.*M_PI/tmpW,0.1).get_shared_plane(0).data(),err);
    int pmax=1<<(n0h-nDecomp+nInit), qmax=1<<(n0w-nDecomp+nInit);
    //-compute coeffs
    //---------------
    if (!divFree){
        CImg<double> tmpUV=fracIntOrDeriv(UV0,zeta)/c;
        cU = wav.anisDecompose2D(tmpUV.get_shared_plane(0), nDecomp)/sqrt(dimW*dimH);
        cV = wav.anisDecompose2D(tmpUV.get_shared_plane(1), nDecomp)/sqrt(dimW*dimH);
        
        
    }
    else{
		CImg<double> tmpUV;
		if (fraction) {
            cout<<"Sorry,  initialization is not possible for fractional series with Leray projection !!"<<endl;
            cU.fill(0.);cV.fill(0.);
        }
		else{
            tmpUV=fracDerivOrInt_CurlDualAdjoint2Part(UV0,0)/c;
            tmpUV.get_shared_plane(0)=wav.anisDecompose2D(tmpUV.get_shared_plane(0), nDecomp)/sqrt(dimW*dimH) ;
            tmpUV.get_shared_plane(1)=wav.anisDecompose2D(tmpUV.get_shared_plane(1), nDecomp)/sqrt(dimW*dimH);
            cU = normalizeFourierDivFreeAnisWavDual(tmpUV);
        }
    }
    //-fill coeffs vector, nI scales
    //-------------------
    theta.fill(0.);
    cout<<"Init 2^-0 to 2^-"<<qmax-1<<endl;
    for (int q=0; q<qmax;q++){
        for (int p=0; p<pmax;p++) {
            theta(p,q,0) = cU(p,q);
            if (!divFree){  theta(p,q,1) = cV(p,q);}
        }
    }
};
//=================
//------------setUV
//=================

void WavOpticalFlow::setUVFracDivFree(const double *X,
                                      int nC){
    
	CImg<double> temp;
	CImg<double> cUV;
	//-create SHARED 2D images corresponding to size of vector X
	if (!fraction){temp.assign(X,1<<(n0h-nDecomp+nC),1<<(n0w-nDecomp+nC),1,1,true);}
	else{temp.assign(X,1<<(n0h-nDecomp+nC),1<<(n0w-nDecomp+nC),2,1,true);}
    
	//-reset
	if (!fraction){cUV.assign(cU.width(),cU.height(),1);cUV.fill(0.);}
	else{cUV.assign(cU.width(),cU.height(),2);cUV.fill(0.);}
    
	//-copy only the nC scales of it into original size images
	cimg_forXYZ(temp,p,q,r){		cUV(p,q,r) = temp(p,q,r);	}
    
	//----------Rebuild UV
	//--------------------
	if (fraction){
		UV.get_shared_plane(0)=wav.anisCompose2D(cUV.get_shared_plane(0),nDecomp)*sqrt(dimW*dimH);
		UV.get_shared_plane(1)=wav.anisCompose2D(cUV.get_shared_plane(1),nDecomp)*sqrt(dimW*dimH);
		UV=c*fracDerivOrInt(UV,zeta,LerayProj);
 	}
	else{
        if (divFree){
            UV.get_shared_plane(0)=wav.anisCompose2D(normalizeFourierDivFreeAnisWav(cUV.get_shared_plane(0)),nDecomp)*sqrt(dimW*dimH);
            UV=c*fracDerivOrInt_Curl(UV.get_shared_plane(0),0);
        }
	}
};

//=================
//----eval_JobsGrad
//=================

double WavOpticalFlow::eval_JobsGradFracDivFree(double *g,const double *x){
	
	double fx = 0;
    
	CImg<double> Ir, Ix, Iy, It,cUVgrad;
	int z = 0;
	int pMax = 1<<(n0h-nDecomp+currentLvl);
	int qMax = 1<<(n0w-nDecomp+currentLvl);
	
	(periodic)? Recal_spline(Ir,I1,UV,3,1) : Recal_spline(Ir,I1,UV,3);
	Ix.assign(gradSpline_X(Ir, sigma, periodic));
	Iy.assign(gradSpline_Y(Ir, sigma, periodic));
	It.assign(Ir-I0);
	
	//---take weights into account
	if (!Weights.is_empty()){
		Ir.assign(It);
		It.mul(Weights);
	}
    //cout << "grad: " << endl;
	//---compute gradient
	cUVgrad.assign(It.width(),It.height(),2);
	cUVgrad.get_shared_plane(0) = It.get_mul(Ix);
	cUVgrad.get_shared_plane(1) = It.get_mul(Iy);
	if (fraction) {
        cUVgrad=c*fracDerivOrInt(cUVgrad,zeta,LerayProj);
        cUVgrad.get_shared_plane(0)=wav.anisDecompose2D(cUVgrad.get_shared_plane(0),nDecomp)/sqrt(dimW*dimH);
		cUVgrad.get_shared_plane(1)=wav.anisDecompose2D(cUVgrad.get_shared_plane(1),nDecomp)/sqrt(dimW*dimH);
    }
	else{
		cUVgrad=c*fracDerivOrInt_CurlAdjoint(cUVgrad,0);
        cUVgrad.get_shared_plane(0)=wav.anisDecompose2D(cUVgrad.get_shared_plane(0),nDecomp)/sqrt(dimW*dimH);
        cUVgrad=normalizeFourierDivFreeAnisWav(cUVgrad);
	}
	for (int q=0; q<qMax; q++) {
		for (int p=0; p<pMax; p++) {
			g[z] = cUVgrad.get_shared_plane(0)(p,q);
			if (fraction){g[nCoeff+z] = cUVgrad.get_shared_plane(1)(p,q);}
			z++;
		}
	}
	//---compute cost fonction
	if (!Weights.is_empty()){
		fx=.5*((It.mul(Ir)).sum());
	}
	else {
		fx=.5*(It.sqr().sum());
	}
	//---return (pour compenser la normalization de la FWT)
	return fx/(dimW*dimH);
};
//=================
//lbfgs_eval_CostFunction
//=================

double WavOpticalFlow::lbfgs_eval_CostFunctionFracDivFree(const double *x,
                                                          double *g){
	
	double fx = 0., fr = 0.;
	
	//---rebuild velocity field
	//-according to current coefficients
	setUVFracDivFree(x, currentLvl);
	
	//---compute Jobs and gradJobs
	fx = eval_JobsGradFracDivFree(g,x);
	
	//---add prior
    if (gamma!=0){
        if (fraction){fr = addFracDivFreeRegul(currentLvl, x, g);;}
        else{
            
            switch (regul) {
                case 1:
                case 2:
                case 3:
                case 4:
                case 5:
                    if ((nDecomp-currentLvl)<4){
                        fr = addHighOrderRegul(x,g);
                    }
                    break;
                case 6:
                    if ((nDecomp-currentLvl)<4){
                        fr=addFracGaussPrior_New(x,g);
                    }
                    break;
                default:
                    break;
            }
        }
    }
    cout<<scientific<<"f:"<<(double)(fx+fr); if (fr!=0){cout<<"="<<scientific<<fx<<"+"<<scientific<<fr;} cout<<" ";
	return fx+fr;
};
//=================
//--------stepSolve
//=================

void WavOpticalFlow::stepSolveFracDivFree(){
    
    int res;
    CImg<double> tu, tv, tu1,tu2, tv1, tv2;
    
    //--number of estimated coefficients (problem size)
    nCoeff = 1<<(n0w + n0h + 2*(currentLvl-nDecomp));
    //--display
    cout << "******************* " <<endl;
    if (currentLvl) {
        cout << "  details lvl [" << n0-nDecomp+currentLvl-1 << "] - ";if (fraction){cout<<"(2x)";}cout<< nCoeff << " coeffs" <<endl;
    }
    else {
        cout << "   coarse lvl [" << n0-nDecomp << "] - ";if (fraction){cout<<"(2x)";}cout<< nCoeff << " coeffs" <<endl;
    }
    cout << "*******************" << endl;
    
    //-load initial guess theta=>deltatheta
    if (!fraction){deltatheta.assign(1<<(n0h-nDecomp+currentLvl),1<<(n0w-nDecomp+currentLvl),1,1, 0.);}
    else{	deltatheta.assign(1<<(n0h-nDecomp+currentLvl),1<<(n0w-nDecomp+currentLvl),2,1, 0.);}
    cimg_forXY(deltatheta,p,q){
        deltatheta(p,q,0) = theta(p,q,0);
        if (fraction){deltatheta(p,q,1) = theta(p,q,1);}
    }
    
    //--go
    paramTheta.max_iterations=150;
    paramTheta.max_linesearch=40;
    paramTheta.m=10;
    paramTheta.linesearch =LBFGS_LINESEARCH_BACKTRACKING_STRONG_WOLFE;//LBFGS_LINESEARCH_BACKTRACKING_STRONG_WOLFE;//LBFGS_LINESEARCH_MORETHUENTE;LBFGS_LINESEARCH_BACKTRACKING_ARMIJO;
    //taux de decroissance minimum de la fct cout
    paramTheta.epsilon=1e-15;
    paramTheta.past=1;
    
    res = lbfgs(deltatheta.size(),deltatheta.data(),&fval,_evaluate,_progress,this,&paramTheta);
    printf("L-BFGS optimization terminated with status code = %d\n", res);
    
    cout << "UPDATE" << endl;
    //-update deltatheta => theta
    cimg_forXY(deltatheta,p,q){
        theta(p,q,0) = deltatheta(p,q,0);
        if (fraction){theta(p,q,1) = deltatheta(p,q,1);}
    }
    //-setUV
    setUVFracDivFree(deltatheta,currentLvl);
    
};

//=================
//--------solve
//=================


CImg<double> WavOpticalFlow::solveFracDivFree(double delta,
                                              const CImg<double> &UV0){
    
    lbfgs_parameter_init(&paramTheta);
    paramTheta.m = 12;
    paramTheta.delta = delta;
    
    
    
    //----------------------------//
    //     initialisations        //
    //----------------------------//
    //----not robust:
    //-use given mask as weights
    Weights.assign(Mask);
    
    //----if first guess given:
    //-initialize UV
    if (!UV0.is_empty()){
        initUVFracDivFree(UV0);
        cout << "++++>> UV initialized with first guess." << endl;
    }
    else{theta(0,0)=1E-14;}
    
    //----------------------------//
    // loop over estimated levels //
    //----------------------------//
    int beginLvl=1;
    double facScale;
    //double gammaFin=gamma;
    if (regul==6){facScale=pow(2.,-1*zeta*2);gamma/=pow(facScale,nDecomp);}
    if (regul==5){facScale=4*4;gamma/=pow(facScale,nDecomp);}
    if (regul==1){facScale=1/*2*2*/;gamma/=pow(facScale,nDecomp);}
	for(currentLvl=beginLvl; currentLvl<=nEstim; currentLvl++) {
        if (!fraction && gamma!=0 ){
            gamma*=facScale;cout<<"gamma="<<gamma<<endl;
            if (regul==6&&approx){
                setFracMatrices_New();
            }
            else{setMatrices();}
        }
        //-step
        stepSolveFracDivFree();
        //output("init",0,0,UV.width(),UV.height());
        //-output visu temp
        if (!UVreal.is_empty()){
            CImg<double> Imvisu(UV.width()*2,UV.height(),2,1,0.);
            CImgList<double> tmp(curlFFT(UV),curlFFT(UVreal));
            cimg_forXY(UV,p,q){Imvisu(p,q,0)=tmp[0](p,q);Imvisu(p+UV.width(),q,0)=tmp[1](p,q);}
            Imvisu.get_shared_plane(0).normalize(0,255).save("toBeErased_curlUV.png");
            if (fraction){
                cimg_forXY(UV,p,q){Imvisu(p,q,0)=theta(p,q,0);Imvisu(p+UV.width(),q,0)=thetaReal(p,q,0);}
                cimg_forXY(UV,p,q){Imvisu(p,q,1)=theta(p,q,1);Imvisu(p+UV.width(),q,1)=thetaReal(p,q,1);}
            }
            else{
                CImg<double> tmpUV1=(fracDerivOrInt(UV,-1*zeta)/c);
                tmpUV1.get_shared_plane(0)=wav.anisDecompose2D(tmpUV1.get_shared_plane(0), nDecomp)/sqrt(dimW*dimH) ;
                tmpUV1.get_shared_plane(1)=wav.anisDecompose2D(tmpUV1.get_shared_plane(1), nDecomp)/sqrt(dimW*dimH) ;
                CImg<double> tmpUV2=(fracDerivOrInt(UVreal,-1*zeta)/c);
                tmpUV2.get_shared_plane(0)=wav.anisDecompose2D(tmpUV2.get_shared_plane(0), nDecomp)/sqrt(dimW*dimH) ;
                tmpUV2.get_shared_plane(1)=wav.anisDecompose2D(tmpUV2.get_shared_plane(1), nDecomp)/sqrt(dimW*dimH) ;
                CImgList<double> tmp2(normalizeFourierDivFreeAnisWavDual(tmpUV1),normalizeFourierDivFreeAnisWavDual(tmpUV2));
                
                cimg_forXYZ(UV,p,q,r){Imvisu(p,q,r)=tmpUV1(p,q,r);Imvisu(p+UV.width(),q,r)=tmpUV2(p,q,r);}
            }
            int err = 0;double N=I0.width();
            double stdevSpec=0.;
            if (!UVreal.is_empty()) {
                write_Spectrum("toBeErased_spectrumTruth.txt",isotropicPSD(UVreal,stdevSpec,2.*M_PI/N,0.1).width(),(isotropicPSD(UVreal,stdevSpec,2.*M_PI/N,0.1).get_shared_plane(0)).data(),err);
            }
            write_Spectrum("toBeErased_spectrum.txt",isotropicPSD(UV,stdevSpec,2.*M_PI/N,0.1).width(),isotropicPSD(UV,stdevSpec,2.*M_PI/N,0.1).get_shared_plane(0).data(),err);
            
        }
    }
    
    //---return
    return UV;
};

//=================
//-stiffnessFractionalMat
//=================
CImg<double> WavOpticalFlow::stiffFracMat_New(float minusZeta, int nLevel, int p, bool b){
    int dim, mid, midCrop, maxVecDim;
    CImg<double> vect,vectCrop, PhiPhi;
    ostringstream namestream;
    
    //-current dimension
    dim = 1<<nLevel;
    
    //-vector's name
    namestream.str("");
    if (b){namestream << coeffDir<<"/w" << wav.get_type() << "m" << wav.get_nMoments() << "z" << fixed<<setprecision(6) << minusZeta << "p" << p << "_.txt";}
    else{namestream << coeffDir<<"/w" << wav.get_type() << "m" << wav.get_nMoments() << "z"<< fixed<<setprecision(6) << minusZeta << "_p" << p << "_.txt";}
    cout << "---------->loading " << namestream.str() << endl;
    vect.assign(const_cast<char*>(namestream.str().c_str()));
    //get rid of zero values
    double temp = vect.get_abs().max();
    cimg_forX(vect, p){
        if ( abs(vect(p)/temp)<1.e-15 ){ vect(p) = 0.; }
    }
    //-fill matrix
    mid = (vect.size()-1)/2;
    maxVecDim=vect.size();
    vectCrop=vect;
    if (dim < vect.size()){ maxVecDim=dim; vectCrop=vect.get_crop(0,mid-dim/2.+1,0,mid-dim/2.+dim);}
    midCrop = (vectCrop.size()-1)/2;
    PhiPhi.assign(dim,dim,1,1, 0.);
    for (int y=0; y<dim; y++) {
        for (unsigned int x=0; x<maxVecDim; x++) {
            PhiPhi(period(x-midCrop+y,dim),y) += vectCrop(x);
        }
    }
    //-finally decompose !
    return wav.anisDecompose2D(PhiPhi, currentLvl);
};

//=================
//------setMatrices
//=================
void WavOpticalFlow::setFracMatrices_New(){
    
    //number of non negelected elements in the newton binomial expansion
    int nb=0;
    double H=-1*zeta-1;
    while (((int)H)+2  >= nb){nb++;}
    cout<<"H="<<H<<", nb="<<nb<<endl;
    
    pStiffW.assign(nb+1);
    pStiffH.assign(nb+1);
    a_pStiffW.assign(nb+1);
    a_pStiffH.assign(nb+1);
    
    
    //-then load matrices
    //-------------
    int i=0;
    while (((int)H)+2  >= i){
        //cout<<2*i<<endl;
        //Phi d^p/dx^p Phi
        pStiffW(i).assign(stiffFracMat_New(-1*zetaIn,n0w-nDecomp+currentLvl,2*i,1));
        pStiffH(i).assign(pStiffW(i).get_transpose());//pStiffW(i).display("p i");
        //cout<<-1*zeta-2*i<<endl;
        //Phi d^(-zeta-p)/dx^(-zeta-p) Phi
        a_pStiffW(i).assign(stiffFracMat_New(-1*zetaIn,n0w-nDecomp+currentLvl,2*i,0));
        a_pStiffH(i).assign(a_pStiffW(i).get_transpose());//a_pStiffW(i).display("a_p i");
        
        i++;
    }
    
    //binary map defining the prior scale range
    setLambda();
};

//=================
//addFracPrior on std basis
//=================

void WavOpticalFlow::setLambda(){
    
	bool normInRange, normXInRange,normYInRange;
	double scaleNorm,scaleX,scaleY,boundx,boundy;
	Lambda.assign(dimW,dimH,1,1);
	Lambda.fill(0.);
	for (int sx=0;sx<log2(Lambda.width());sx++){
        for (int sy=0;sy<log2(Lambda.height());sy++){
            scaleNorm=sqrt(pow(2.,2*(n0w-sx))+pow(2.,2*(n0h-sy)));
            scaleX=pow(2.,(n0w-sx-1));
            scaleY=pow(2.,(n0h-sy-1));
            normInRange=(scaleNorm <= MaxPwrRange && scaleNorm >= MinPwrRange);
            //normXInRange=(scaleX <= MaxPwrRange && scaleX >= MinPwrRange);
            //normYInRange=(scaleY <= MaxPwrRange && scaleY >= MinPwrRange);
            if (normInRange /*|| normXInRange || normYInRange*/){
                if (sx==0){boundx=0;}else{boundx=(1<<sx);}
                if (sy==0){boundy=0;}else{boundy=(1<<sy);}
                for (int px=boundx;px<(1<<(sx+1));px++){
                    for (int py=boundy;py<(1<<(sy+1));py++){
                        Lambda(px,py) =1;
                    }
                }
            }
        }
	}
 	Lambda.get_normalize(0,255).save_png("toBeErased_Lambda.png");
    //Lambda.display();
}

double WavOpticalFlow::addFracGaussPrior_New(const double *x,
                                             double *g){
    
    cout<<"***";
    CImg<double> gradtemp,funcTmp, coefftemp, LambdaPart, App,AppTmp;
    double i=0;
    int z=0;
    int nC=(n0w-nDecomp+currentLvl);
    //-create SHARED 2D images corresponding to size of vector X
    coefftemp.assign(x,1<<(n0h-nDecomp+currentLvl),1<<(n0w-nDecomp+currentLvl),1,1,true);
    AppTmp.assign(1<<(n0h-nDecomp+currentLvl),1<<(n0w-nDecomp+currentLvl),1,1,false);
    
    //-create NONSHARED 2D images corresponding to size of vector X
    gradtemp.assign(coefftemp,false);
    gradtemp.fill(0.);
    funcTmp.assign(coefftemp,false);
    funcTmp.fill(0.);
    LambdaPart=Lambda.get_crop(0,0,(1<<(n0h-nDecomp+currentLvl))-1,(1<<(n0w-nDecomp+currentLvl))-1);
    
    //-----------------------------------------------------
    //calculus of gradient and objective function
    //wavelet coefficients of motion components in a fractional basis
    //contre normalization.
    double H=-1*zeta-1;
    
    if (!approx){//(no tensorial form)
        cout<<"*";
        AppTmp=fracDerivOrInt(wav.anisCompose2D(coefftemp,nC),2*(H+2))/dimH;
        gradtemp= wav.anisDecompose2D(AppTmp, nC)/dimH;
    }
    else{//(tensorial form)
        
        i=0;
        AppTmp.fill(0);
        double factorNorm=pow(2.,n0w+2*(currentLvl-nDecomp));
        while (((int)H)+2  >= i){//symetrized
            AppTmp+=factorNorm*(a_pStiffH(i)*coefftemp*pStiffW(i))/2/dimH;
            AppTmp+=factorNorm*(pStiffH(i)*coefftemp*a_pStiffW(i))/2/dimH;
            i++;
        }
        gradtemp=AppTmp;
    }
    funcTmp=0.5*coefftemp.get_mul(gradtemp);
    //-add it to global gradient
    z=0;
    cimg_forXY(gradtemp,p,q){
        g[z] += gamma*gradtemp(p,q,0);
        z++;
    }
    //-and compute the sum
    return gamma*funcTmp.sum();
    
};

//=================
//addFracPrior on Fractional basis
//=================

double WavOpticalFlow::addFracDivFreeRegul(int lvl,
                                           const double *x,
                                           double *g){
    cout<<"+++";
	//carte poids regularization
	CImg<double> gammaMap(dimW,dimH,1,1);gammaMap.fill(gamma);
	gammaMap.crop(0,0,(1<<lvl)-1,(1<<lvl)-1);
    
	
	int boundx,boundy;
	CImg<double> tempX;
	tempX.assign(x,1<<lvl,1<<lvl,2,1,true);
	CImg<double> tempG;
	tempG.assign(g,1<<lvl,1<<lvl,2,1,true);
	double freg = 0.;
	double scaleNorm;
	for (int sx=0;sx<log2(tempX.width());sx++){
        for (int sy=0;sy<log2(tempX.height());sy++){
            scaleNorm=sqrt(pow(2.,2*(n0w-sx-1))+pow(2.,2*(n0h-sy-1)));
            if (scaleNorm <= MaxPwrRange && scaleNorm >= MinPwrRange){
                if (sx==0){boundx=0;}else{boundx=(1<<sx);}
                if (sy==0){boundy=0;}else{boundy=(1<<sy);}
                for (int px=boundx;px<(1<<(sx+1));px++){
                    for (int py=boundy;py<(1<<(sy+1));py++){
                        tempG(px,py,0) += gammaMap(px,py)*tempX(px,py,0);
                        tempG(px,py,1) += gammaMap(px,py)*tempX(px,py,1);
                        freg += gammaMap(px,py)*tempX(px,py,0)*tempX(px,py,0);
                        freg += gammaMap(px,py)*tempX(px,py,1)*tempX(px,py,1);
                    }
                }
            }
        }
	}

	return .5*freg;
};



/********************************************************************************************************************************************
 *
 *  WavOpticalFlow class
 *
 *******************************************************************************************************************************************/

//=================
//------constructor
//=================
WavOpticalFlow::WavOpticalFlow(CImg<double> Im0,
                               CImg<double> Im1,
                               CImg<double> ImMask,
                               int nD,
                               int nE,
                               int nI,
                               int nM,
                               int wT,
                               double sig,
                               int reg,
                               double a) :
I0(Im0), I1(Im1), UV(Im0.width(),Im0.height(),2,1,0.), Mask(ImMask),
dimW(Im0.width()), dimH(Im0.height()), nDecomp(nD), nEstim(nE), nInit(nI), currentLvl(0), wav(nM,wT),
regul(reg), sigma(sig)
{
    //-allocation wavelet coeff /!\ transposed dimensions !!
    cU.assign(I0.height(),I0.width(),1,1, 0.);
    cV.assign(I0.height(),I0.width(),1,1, 0.);
    //-then we can also check & correct nEstim value
    if (nEstim>nDecomp) {
        cout << "\t /!\\ Impossible estimation requested (" << nEstim <<"), corrected to " << nDecomp << "(full resolution estimation)"<< endl;
        nEstim = nDecomp;
    }
    else if (nEstim<0){
        cout << "\t /!\\ Impossible estimation requested (" << nEstim <<" levels), corrected to " << 0 << "(coarse only)"<< endl;
        nEstim = 0;
    }
    
    //-check mask, remove if not necessary.
    if ((Mask.min()==1.)&&(Mask.max()==1.)) {Mask.assign();}
    
    //-regularization weights
    gamma = a;
    
    //-original scale
    n0w = (int)log2(dimW);
    n0h = (int)log2(dimH);
    n0 = min(n0w,n0h);
    
    //-total number of coeff
    nTotCoeff = 1<<(n0w+n0h+2*(nEstim-nDecomp));
    //-initialize theta
    theta.assign(1<<(n0h-nDecomp+nEstim),1<<(n0w-nDecomp+nEstim),2,1, 0.);
    divFree=false;
    fraction=false ;
    
};


//=================
//-------destructor
//=================
WavOpticalFlow::~WavOpticalFlow(){
    
    cout << "=> WavOpticalFlow object killed ..."<<endl;
};

//=================
//-------------info
//=================
void WavOpticalFlow::info(bool disp){
    
    cout << "----------------------------------" << endl;
    cout << "dimensions: " << dimW << "x" << dimH <<endl;
    cout << "niveaux d'approximation original:" << n0w << "x" << n0h << endl;
    cout << "niveaux de decomposition: " << nDecomp <<endl;
    cout << "niveaux estimes: " << nEstim << endl;
    cout << "sigma: " << sigma <<endl;
    cout << "wavelet: ";
    wav.info();
    cout << "divergence-free basis: "<<divFree<<endl;
    cout << "regularisation: ";
    switch (regul) {
        case 0:
            cout << "" << endl;
            break;
        case 1:
            cout << "order 1 regularization" << endl;
            break;
        case 2:
            cout << "order 2 regularization" << endl;
            break;
        case 3:
            cout << "laplacian regularization" << endl;
            break;
        case 4:
            cout << "div regularization" << endl;
            break;
        case 5:
            cout << "gradcurl regularization" << endl;
            break;
        case 6:
            cout << "fractional regularization" << endl;
            break;
        case -1:
            cout << "none" << endl;
            break;
    }
    if ((regul>0)&&(regul<=9)) {cout << "gamma: " << gamma << endl;}
    cout << "border condition: ";
    (periodic)? cout << "periodic" : cout << "Neumann-0";
    cout << endl;
    cout << "----------------------------------" << endl;
    if (disp) CImgList<double>(I0,I1,I1-I0,true).display();
};

//=================
//-----------initUV
//=================
void WavOpticalFlow::initUV(const CImg<double> &UV0){
    
    int pmax=1<<(n0h-nDecomp+nInit), qmax=1<<(n0w-nDecomp+nInit);
    
    //-compute coeffs
    //---------------
    cU = wav.anisDecompose2D(UV0.get_shared_plane(0), nDecomp)/sqrt(dimW*dimH);
    cV = wav.anisDecompose2D(UV0.get_shared_plane(1), nDecomp)/sqrt(dimW*dimH);
    //-fill coeffs vector, nI scales
    //-------------------
    theta.fill(0.);
    for (int q=0; q<qmax;q++){
        for (int p=0; p<pmax;p++) {
            theta(p,q,0) = cU(p,q);
            theta(p,q,1) = cV(p,q);
        }
    }
};

//=================
//------------setUV
//=================
void WavOpticalFlow::setUV(const double *X,
                           int nC){
    
    CImg<double> temp;
    
    //-create SHARED 2D images corresponding to size of vector X
    temp.assign(X,1<<(n0h-nDecomp+nC),1<<(n0w-nDecomp+nC),2,1,true);
    
    //-reset
    cU.fill(0.);
    cV.fill(0.);
    //-copy only the nC scales of it into original size images
    cimg_forXY(temp,p,q){
        cU(p,q) = temp(p,q,0);
        cV(p,q) = temp(p,q,1);
    }
    //----------Rebuild UV
    //--------------------
    UV.get_shared_plane(0) = wav.anisCompose2D(cU, nDecomp)*sqrt(dimW*dimH);
    UV.get_shared_plane(1) = wav.anisCompose2D(cV, nDecomp)*sqrt(dimW*dimH);
    
};


//=================
//--------eval_Jobs
//=================
double WavOpticalFlow::eval_Jobs(){
    
    CImg<double> Ir;
    //-recal
    (periodic)? Recal_spline(Ir,I1,UV,3,1) : Recal_spline(Ir,I1,UV,3);
    //-difference
    Ir-=I0;
    //-use weights if any
    if (!Weights.is_empty()){Ir.mul(Weights);}
    return .5*((Ir.sqr()).sum())/(Ir.width()*Ir.height());
};

//=================
//----eval_JobsGrad
//=================
double WavOpticalFlow::eval_JobsGrad(double *g){
    
    double fx = 0;
    CImg<double> Ir, Ix, Iy, It, Itemp, tempgradU, tempgradV;
    int z = 0;
    int pMax = 1<<(n0h-nDecomp+currentLvl);
    int qMax = 1<<(n0w-nDecomp+currentLvl);
    
    (periodic)? Recal_spline(Ir,I1,UV,3,1) : Recal_spline(Ir,I1,UV,3);
    Ix.assign(gradSpline_X(Ir, sigma));
    Iy.assign(gradSpline_Y(Ir, sigma));
    It.assign(Ir-I0);
    
    //---take weights into account
    if (!Weights.is_empty()){
        Ir.assign(It);
        It.mul(Weights);
    }
    //---compute gradient
    Itemp = It.get_mul(Ix);
    tempgradU.assign(wav.anisDecompose2D(Itemp, nDecomp)/sqrt(dimW*dimH));
    Itemp = It.get_mul(Iy);
    tempgradV.assign(wav.anisDecompose2D(Itemp, nDecomp)/sqrt(dimW*dimH));
    for (int q=0; q<qMax; q++) {
        for (int p=0; p<pMax; p++) {
            g[z] = tempgradU(p,q);
            g[nCoeff+z] = tempgradV(p,q);
            z++;
        }
    }
    //---compute cost fonction
    if (!Weights.is_empty()){
        fx=.5*((It.mul(Ir)).sum());
    }
    else {
        fx=.5*(It.sqr().sum());
    }
    //---return (pour compenser normalization de FWT)
    return fx/(It.width()*It.height());
};

//=================
//lbfgs_eval_CostFunction
//=================
double WavOpticalFlow::lbfgs_eval_CostFunction(const double *x,
                                               double *g){
    
    double fx = 0., fr = 0.;
    
    //---rebuild velocity field
    //-according to current coefficients
    setUV(x, currentLvl);
    //---compute Jobs and gradJobs
    fx = eval_JobsGrad(g);
    
    //-regul
    switch (regul) {
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
            if (true){
                fr = addHighOrderRegul(x,g);
                cout << "fx: " << fx << " ; fr: " << fr << endl;
            }
            break;
        default:
            break;
    }
    
    return fx+fr;
};
//=================
//addHighOrderRegul
//=================
double WavOpticalFlow::addHighOrderRegul(const double *x,
                                         double *g){
    
    CImg<double> gradtemp, coefftemp;
    double res = 0.;
    int z=0,zRef=0;
    int divLabel=0;
    if (!fraction){divLabel+=10;
        if (divFree){divLabel+=10;}
    }
    
    
    //-create SHARED 2D images corresponding to size of vector X
    coefftemp.assign(x,1<<(n0h-nDecomp+currentLvl),1<<(n0w-nDecomp+currentLvl),2,1,true);
    cout<<"--- "<<regul+divLabel;
    switch (regul+divLabel) {
        case 21:
            // div free case
            //-regul = |dx(u)|^2 + |dy(u)|^2 + |dx(v)|^2 +|dy(v)|^2
            //       = |dxx(chi)|^2 + |dyy(chi)|^2 + |2*dxdy(chi)|^2
            //-----------------------------------------------------
            //-compute gradient
            //|dx(chi)|^2
            gradtemp  = (evenMass(2)*coefftemp.get_shared_plane(0)*evenMass(0));
            //|dy(chi)|^2
            gradtemp += (evenMass(0)*coefftemp.get_shared_plane(0)*evenMass(2));
            //|2*dxdy(chi)|^2
            gradtemp += (evenMass(1)*coefftemp.get_shared_plane(0)*evenMass(1));
            //-add it to global gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            //-and compute the l2norm
            res += (gradtemp.mul(coefftemp.get_shared_plane(0))).sum();
            break;
        case 25:
            // div free case
            //-regul = |gradcurl(u,v)|^2 = |Laplacian(u)|^2 + |Laplacian(v)|^2 =  |grad Laplacian(chi)|^2
            //       = |dxxx(chi)|^2 + |dyyy(chi)|^2 + 3*|dxxdy(chi)|^2 + 3*|dxdyy(chi)|^2
            //-----------------------------------------------------
            //-compute gradient
            //|dxxx(chi)|^2
            //if(currentLvl>5){ evenMass(2).display("evenmass2 flumitools"); evenMass(3).display("evenmass3 flumitools");}
            gradtemp  = (evenMass(3)*coefftemp.get_shared_plane(0)*evenMass(0));
            //|dyyy(chi)|^2
            gradtemp += (evenMass(0)*coefftemp.get_shared_plane(0)*evenMass(3));
            //|dxxdy(chi)|^2
            gradtemp += 3*(evenMass(2)*coefftemp.get_shared_plane(0)*evenMass(1));
            //|dxdyy(chi)|^2
            gradtemp += 3*(evenMass(1)*coefftemp.get_shared_plane(0)*evenMass(2));
            //-add it to global gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            //-and compute the l2norm
            res += (gradtemp.mul(coefftemp.get_shared_plane(0))).sum();
            break;
        case 11:
            //-regul = |dx(u)|^2 + |dy(u)|^2 + |dx(v)|^2 +|dy(v)|^2
            //-----------------------------------------------------
            //-compute gradient for U term
            gradtemp = (evenMass(1)*coefftemp.get_shared_plane(0)*evenMass(0)) + (evenMass(0)*coefftemp.get_shared_plane(0)*evenMass(1));
            //-add it to global gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            //-and compute the l2norm
            res += (gradtemp.mul(coefftemp.get_shared_plane(0))).sum();
            //-compute gradient for V term
            gradtemp = (evenMass(1)*coefftemp.get_shared_plane(1)*evenMass(0)) + (evenMass(0)*coefftemp.get_shared_plane(1)*evenMass(1));
            //-add it to global gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            //-and compute the l2norm
            res += (gradtemp.mul(coefftemp.get_shared_plane(1))).sum();
            break;
        case 12:
            //-regul = |dxx(u)|^2 + |dyy(u)|^2 + |dxx(v)|^2 + |dyy(v)|^2
            //----------------------------------------------------------
            //-compute gradient for U term
            gradtemp = (evenMass(2)*coefftemp.get_shared_plane(0)*evenMass(0)) + (evenMass(0)*coefftemp.get_shared_plane(0)*evenMass(2));
            //-add it to global gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            //-and compute the l2norm
            res += (gradtemp.mul(coefftemp.get_shared_plane(0))).sum();
            //-compute gradient for V term
            gradtemp = (evenMass(2)*coefftemp.get_shared_plane(1)*evenMass(0)) + (evenMass(0)*coefftemp.get_shared_plane(1)*evenMass(2));
            //-add it to global gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            //-and compute the l2norm
            res += (gradtemp.mul(coefftemp.get_shared_plane(1))).sum();
            break;
        case 13:
            //-regul = |Laplacian(u)|^2 + |Laplacian(v)|^2
            //----------------------------------
            //-compute gradient for U term
            gradtemp = (evenMass(2)*coefftemp.get_shared_plane(0)*evenMass(0)) + (evenMass(0)*coefftemp.get_shared_plane(0)*evenMass(2)) + 2.*(evenMass(1)*coefftemp.get_shared_plane(0)*evenMass(1));
            //-add it to global gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            //-and compute the l2norm
            res += (gradtemp.mul(coefftemp.get_shared_plane(0))).sum();
            //-compute gradient for V term
            gradtemp = (evenMass(2)*coefftemp.get_shared_plane(1)*evenMass(0)) + (evenMass(0)*coefftemp.get_shared_plane(1)*evenMass(2)) + 2.*(evenMass(1)*coefftemp.get_shared_plane(1)*evenMass(1));
            //-add it to global gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            //-and compute the l2norm
            res += (gradtemp.mul(coefftemp.get_shared_plane(1))).sum();
            break;
        case 14:
            //-regul = |div(u,v)|^2
            //-      = (dxU)^2 + (dyV)^2 + 2(dxUdyV)
            //----------------------------------
            zRef=0;
            //-compute gradient for U:
            gradtemp = (evenMass(1)*coefftemp.get_shared_plane(0)*evenMass(0));
            //-add contrib to cost function and to main gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            res += (gradtemp.mul(coefftemp.get_shared_plane(0))).sum();
            //-compute other part (mixed U,V)
            gradtemp = (oddMass(0)*coefftemp.get_shared_plane(1)*oddMass(0));
            z=zRef;
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            res += 2*(gradtemp.mul(coefftemp.get_shared_plane(0))).sum();
            zRef=z;
            //-compute gradient for V:
            gradtemp = (evenMass(0)*coefftemp.get_shared_plane(1)*evenMass(1));
            //-add contrib to cost function and to main gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            res += (gradtemp.mul(coefftemp.get_shared_plane(1))).sum();
            //-the remaning part of gradient
            gradtemp = (oddMass(1)*coefftemp.get_shared_plane(0)*oddMass(1));
            z=zRef;
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            break;
        case 15:
            //-regul = |gradcurl(u,v)|^2
            //-  = (dyyU)^2 + (dxyU)^2 + (dxxV)^2 + (dyxV)^2 - 2[ (dxyU)(dxxV) + (dyyU)(dyxV) ]
            //----------------------------------
            zRef=0;
            //-compute half gradient for U term
            gradtemp = (evenMass(0)*coefftemp.get_shared_plane(0)*evenMass(2)) + (evenMass(1)*coefftemp.get_shared_plane(0)*evenMass(1));
            //-add contrib to cost function and to main gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            res += (gradtemp.mul(coefftemp.get_shared_plane(0))).sum();
            //-compute the other half (mixed u/v)
            gradtemp = - (oddMass(2)*coefftemp.get_shared_plane(1)*oddMass(0)) - (oddMass(1)*coefftemp.get_shared_plane(1)*oddMass(3));
            z=zRef;
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            res += 2.*((gradtemp.mul(coefftemp.get_shared_plane(0))).sum());
            zRef=z;
            //-compute half gradient for V term
            gradtemp = (evenMass(2)*coefftemp.get_shared_plane(1)*evenMass(0)) + (evenMass(1)*coefftemp.get_shared_plane(1)*evenMass(1));
            //-add contrib to cost function and main gradient
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            res += (gradtemp.mul(coefftemp.get_shared_plane(1))).sum();
            //-compute the other half of gradient (mix u/v)
            //-REMARK here we must not add it to cost function, it has already been done above with u.
            gradtemp = - (oddMass(3)*coefftemp.get_shared_plane(0)*oddMass(1)) - (oddMass(0)*coefftemp.get_shared_plane(0)*oddMass(2));
            //-add it to global gradient
            z=zRef;
            cimg_forXY(gradtemp,p,q){
                g[z] += gamma*gradtemp(p,q);
                z++;
            }
            break;
        default:
            cout<<"carefull, no regularization label correspondance ... "<<endl;
            break;
    }
    
    return .5*gamma*res;
};

//=================
//---------progress
//=================
int WavOpticalFlow::progress(const double *x,
                             const double *g,
                             const double fx,
                             const double xnorm,
                             const double gnorm,
                             const double step,
                             int n,
                             int k,
                             int ls){
    
    printf("Iteration %d:\n", k);
    printf("  fx = %e", fx);
    printf("  xnorm = %e, gnorm = %e, step = %e\n", xnorm, gnorm, step);
    printf("\n");
    return 0;
};

//=================
//--------stepSolve
//=================
void WavOpticalFlow::stepSolve(){
    
    int res;
    CImg<double> tu, tv, tu1,tu2, tv1, tv2;
    
    //--number of estimated coefficients (problem size)
    nCoeff = 1<<(n0w + n0h + 2*(currentLvl-nDecomp));
    //--display
    cout << "*******************" << endl;
    if (currentLvl) {
        cout << "  details lvl [" << n0-nDecomp+currentLvl-1 << "] - (2x)"<< nCoeff << " coeffs" <<endl;
    }
    else {
        cout << "   coarse lvl [" << n0-nDecomp << "] - (2x)"<< nCoeff << " coeffs" <<endl;
    }
    cout << "*******************" << endl;
    
    //-load initial guess theta=>deltatheta
    deltatheta.assign(1<<(n0h-nDecomp+currentLvl),1<<(n0w-nDecomp+currentLvl),2,1, 0.);
    cimg_forXY(deltatheta,p,q){
        deltatheta(p,q,0) = theta(p,q,0);
        deltatheta(p,q,1) = theta(p,q,1);
    }
    
    //--go
    res = lbfgs(deltatheta.size(),deltatheta.data(),&fval,_evaluate,_progress,this,&paramTheta);
    printf("L-BFGS optimization terminated with status code = %d\n", res);
    cout << "UPDATE" << endl;
    //-update deltatheta => theta
    cimg_forXY(deltatheta,p,q){
        theta(p,q,0) = deltatheta(p,q,0);
        theta(p,q,1) = deltatheta(p,q,1);
    }
    //-setUV
    setUV(deltatheta,currentLvl);
};


//=================
//------------solve
//=================
CImg<double> WavOpticalFlow::solve(double delta,
                                   const CImg<double> &UV0,
                                   bool interp,
                                   bool div){
    
    lbfgs_parameter_init(&paramTheta);
    paramTheta.m = 12;
    paramTheta.delta = delta;
    
    //----------------------------//
    //     initialisations        //
    //----------------------------//
    
    //----not robust:
    //-use given mask as weights
    Weights.assign(Mask);
    
    //----if first guess given:
    //-initialize UV
    if (!UV0.is_empty()){
        initUV(UV0);
        cout << "++++>> UV initialized with first guess." << endl;
    }
    
    //----------------------------//
    // loop over estimated levels //
    //----------------------------//
    for(currentLvl=0; currentLvl<=nEstim; currentLvl++) {
        //-these matrices correspond to the current resolution space
        //-therefore they need to be re-computed at each new scale.
        if ((regul>0)){
            setMatrices();
            
        }
        //-step
        stepSolve();
        
        int err = 0;double N=I0.width();double stdevSpec=0.;
        write_Spectrum("toBeErased_spectrum.txt",isotropicPSD(UV,stdevSpec,2.*M_PI/N,0.1).width(),isotropicPSD(UV,stdevSpec,2.*M_PI/N,0.1).get_shared_plane(0).data(),err);
        
        
        
    }
    
    //---return
    return UV;
};

//=================
//----------compare
//=================
void WavOpticalFlow::compare(CImg<double>& trueField,
                             bool disp){
    
    double err, errB;
    int excludeMargin = 30;
    
    //--display
    cout << "*******************" << endl;
    cout << "  Comparison !" << endl;
    cout << "*******************" << endl;
    
    
    if (disp){
        if (!Mask.is_empty()){CImgList<double>(trueField.get_shared_plane(0), UV.get_shared_plane(0).get_mul(Mask), trueField.get_shared_plane(1), UV.get_shared_plane(1).get_mul(Mask), true).display("compare: U truth, U estim, V truth, V estim");}
        else {CImgList<double>(trueField.get_shared_plane(0), UV.get_shared_plane(0), trueField.get_shared_plane(1), UV.get_shared_plane(1), true).display("compare: U truth, U estim, V truth, V estim");}
        divFFT(UV).display("div UV");curlFFT(UV).display("curl UV");
    }
    CImg<double> errU = (trueField.get_shared_plane(0)-UV.get_shared_plane(0)).sqr();
    CImg<double> errV = (trueField.get_shared_plane(1)-UV.get_shared_plane(1)).sqr();
    errEndPoint = errU+errV;
    errBarron = ((1+UV.get_shared_plane(0).get_mul(trueField.get_shared_plane(0))+UV.get_shared_plane(1).get_mul(trueField.get_shared_plane(1))).div((1+UV.get_shared_plane(0).get_mul(UV.get_shared_plane(0))+UV.get_shared_plane(1).get_mul(UV.get_shared_plane(1))).sqrt().get_mul((1+trueField.get_shared_plane(0).get_mul(trueField.get_shared_plane(0))+trueField.get_shared_plane(1).get_mul(trueField.get_shared_plane(1))).sqrt()))).acos()*180/M_PI;
    if (!Mask.is_empty()){
        errBarron.mul(Mask);
        errEndPoint.mul(Mask);
    }
    else {
        errBarron.crop(excludeMargin,excludeMargin,dimW-excludeMargin-1,dimH-excludeMargin-1);
        errEndPoint.crop(excludeMargin,excludeMargin,dimW-excludeMargin-1,dimH-excludeMargin-1);
    }
    err = sqrt(errEndPoint.mean());
    errB = errBarron.mean();
    errEndPoint.sqrt();
    cout << "-->>> error: " << err << " .. " << errB << endl;
    
    //--trace
    fstream frez("logResults.txt", ios::out | ios::app);
    if (frez){
        frez << wav.get_nMoments() << "\t" << nDecomp << "\t" << nEstim << "\t" << sigma << "\t" << regul << "\t" << gamma << "\t" << err << "\t" << errB << endl;
    }
    
    if (disp) errEndPoint.display("Vector norm error on  UV");
    if (disp) errBarron.display("Barron error on  UV");
    
};

//=================
//-----------output
//=================
void WavOpticalFlow::output(string outDir, int Ox, int Oy, int dimx, int dimy){
    
    CImg<double>UVout = UV.get_crop(Ox,Oy,Ox+dimx-1,Oy+dimy-1);
    saveMotion(const_cast<char*>((outDir + "UV.inr").c_str()), UVout);
    saveMotion(const_cast<char*>((outDir + "UV.motion").c_str()), UVout);

 
    
};

//=================
//----------massMat
//=================
CImg<double> WavOpticalFlow::massMat(int order, int nLevel){
    int dim, mid;
    CImg<double> vect, PhiPhi;
    ostringstream namestream;
    
    //-current dimension
    dim = 1<<nLevel;
    
    //-vector's name
    namestream.str("");
    namestream << coeffDir<<"/w" << wav.get_type() << "m" << wav.get_nMoments() << "p" << order << ".txt";
    cout << "---------->loading " << namestream.str() << endl;
    vect.assign(const_cast<char*>(namestream.str().c_str()));
    //-factor to get to current resolution values
    vect *= pow(2.,order*(nLevel-n0w));
    //-fill matrix
    mid = (vect.size()-1)/2;
    PhiPhi.assign(dim,dim,1,1, 0.);
    for (int y=0; y<dim; y++) {
        for (unsigned int x=0; x<vect.size(); x++) {
            PhiPhi(period(x-mid+y,dim),y) += vect(x);
        }
    }
    //-finally decompose
    return wav.anisDecompose2D(PhiPhi, currentLvl)*sqrt((double)PhiPhi.size());
};


//=================
//------setMatrices
//=================
void WavOpticalFlow::setMatrices(){
    
    bool loadM0=false, loadM1=false, loadM2=false, loadM3=false, loadM4=false, loadM6=false;
    
    evenMass.assign(4);
    oddMass.assign(4);
    
    int divLabel=0;
    if (!fraction){divLabel+=10;
        if (divFree){divLabel+=10;}
    }
    //--here we define which matrices are to be loaded
    //-according to the chosen regularization
    switch (regul+divLabel) {
            //-Horn & Schunk
        case 11:
            loadM0 = true;
            loadM2 = true;
            break;
        case 21: //div free
            loadM0 = true;
            loadM2 = true;
            loadM4 = true;
            break;
            //-order 2
        case 12:
            loadM0 = true;
            loadM4 = true;
            break;
            //-laplacian
        case 13:
            loadM0 = true;
            loadM2 = true;
            loadM4 = true;
            break;
            //-divergence
        case 14:
            loadM0 = true;
            loadM1 = true;
            loadM2 = true;
            break;
            //-gradcurl
        case 15:
            loadM0 = true;
            loadM1 = true;
            loadM2 = true;
            loadM3 = true;
            loadM4 = true;
            break;
        case 25: //div free
            loadM0 = true;
            loadM2 = true;
            loadM4 = true;
            loadM6 = true;
            break;
        default:
            break;
    }
    
    //-then load it
    //-------------
    if (loadM0) {
        //-PhiPhi
        evenMass(0).assign(massMat(0,n0w-nDecomp+currentLvl));
    }
    if (loadM1){
        //-PhiPhi'
        oddMass(0).assign(massMat(1,n0w-nDecomp+currentLvl));
        //-Phi'Phi (transposed of previous)
        oddMass(1).assign(oddMass(0).get_transpose());
    }
    if (loadM2){
        //-Phi'Phi'
        evenMass(1).assign(massMat(2,n0w-nDecomp+currentLvl));
    }
    if (loadM3){
        //-Phi'Phi''
        oddMass(2).assign(massMat(3,n0w-nDecomp+currentLvl));
        //-Phi''Phi' (transposed of previous)
        oddMass(3).assign(oddMass(2).get_transpose());
    }
    if (loadM4){
        //-Phi''Phi''
        evenMass(2).assign(massMat(4,n0w-nDecomp+currentLvl));
    }
    if (loadM6){
        //-Phi'''Phi'''
        evenMass(3).assign(massMat(6,n0w-nDecomp+currentLvl));
    }
    
};

/*********************
 **********************
 **                  **
 **  Static methods  **
 **                  **
 **********************
 *********************/

//=================
//--prepareInputImg
//=================
void WavOpticalFlow::prepareInputImg(CImg<double> I0in,
                                     CImg<double> I1in,
                                     const CImg<double> &Bgin,
                                     const CImg<double> &Maskin,
                                     const CImg<double> &UV0in,
                                     const CImg<double> &UVTin,
                                     CImg<double> &I0out,
                                     CImg<double> &I1out,
                                     CImg<double> &Maskout,
                                     CImg<double> &UV0out,
                                     CImg<double> &UVTout,  
                                     int &Ox,
                                     int &Oy){
    
    //-get nearest pow 2 dimensions
    //--------------------------
    int Dx = cimg::nearest_pow2((unsigned int)I0in.width());
    int Dy = cimg::nearest_pow2((unsigned int)I0in.height());  
    double minVal, maxVal;
    CImg<double> tempMask;
    //-and offsets
    Ox = (Dx - I0in.width())/2;
    Oy = (Dy - I0in.height())/2;
    
    //-check original dimensions
    //--------------------------
    if (I0in.size()!=I1in.size()) cout << "Haaaaaaaaaaaaaaaa I !!!" << endl;
    if ((!Maskin.is_empty())&&((I0in.size()!=Maskin.size()))) cout << "Haaaaaaaaaaaaaaaa M !!!" << endl;
    if ((!UV0in.is_empty())&&(((int)I0in.size()!=UV0in.width()*UV0in.height()))) cout << "Haaaaaaaaaaaaaaaa UV0 !!!" << endl;
    if ((!UVTin.is_empty())&&(((int)I0in.size()!=UVTin.width()*UVTin.height()))) cout << "Haaaaaaaaaaaaaaaa UVtrue !!!" << endl;
    
    //-create powers-2 pictures
    //-------------------------
    I0out.assign(Dx,Dy,1,1,0.);
    I1out.assign(Dx,Dy,1,1,0.);
    Maskout.assign(Dx,Dy,1,1,0.);
    if (!UV0in.is_empty()) {UV0out.assign(Dx,Dy,2,1,0.);}
    if (!UVTin.is_empty()) {UVTout.assign(Dx,Dy,2,1,0.);}
    
    //-values for normalization
    //-------------------------
    //-first remove background, if one
    if (!Bgin.is_empty()){
        I0in -= Bgin;
        I1in -= Bgin;
    }
    //-if a mask is given, we select min & max values over non-masked data only.
    if (!Maskin.is_empty()) {
        //-also normalize input mask (safety)
        tempMask.assign(Maskin);
        tempMask.normalize();
        minVal = cimg::min((I0in.get_mul(tempMask)).min(),(I1in.get_mul(tempMask)).min());
        maxVal = cimg::max((I0in.get_mul(tempMask)).max(),(I1in.get_mul(tempMask)).max());  
    }
    else {
        minVal = cimg::min(I0in.min(),I1in.min());
        maxVal = cimg::max(I0in.max(),I1in.max());
    }
    maxVal = 1./(maxVal-minVal);
    
    //-paste original and normalize by the way
    cimg_forXY(I0in,x,y){
        I0out(Ox+x,Oy+y) = (I0in(x,y)-minVal)*maxVal;
        I1out(Ox+x,Oy+y) = (I1in(x,y)-minVal)*maxVal;
        Maskout(Ox+x,Oy+y) = (tempMask.is_empty())? 1. : tempMask(x,y);
        if (!UV0out.is_empty()) { 
            UV0out(Ox+x,Oy+y,0) = UV0in(x,y,0);
            UV0out(Ox+x,Oy+y,1) = UV0in(x,y,1);
        }
        if (!UVTout.is_empty()) { 
            UVTout(Ox+x,Oy+y,0) = UVTin(x,y,0);
            UVTout(Ox+x,Oy+y,1) = UVTin(x,y,1);
        }
    }
};

//=================
//--------_evaluate
//=================
lbfgsfloatval_t WavOpticalFlow::_evaluate(void *instance,
                                          const lbfgsfloatval_t *x,
                                          lbfgsfloatval_t *g,
                                          const int n,
                                          const lbfgsfloatval_t step){
    double res;
    if (reinterpret_cast<WavOpticalFlow*>(instance)->fraction||reinterpret_cast<WavOpticalFlow*>(instance)->divFree ){cout<<"_";fflush(stdout);res=reinterpret_cast<WavOpticalFlow*>(instance)->lbfgs_eval_CostFunctionFracDivFree(x, g);}
    else{cout<<".";fflush(stdout);res=reinterpret_cast<WavOpticalFlow*>(instance)->lbfgs_eval_CostFunction(x, g);}
    return res;
};

//=================
//--------_progress
//=================
int WavOpticalFlow::_progress(void *instance,
                              const lbfgsfloatval_t *x,
                              const lbfgsfloatval_t *g,
                              const lbfgsfloatval_t fx,
                              const lbfgsfloatval_t xnorm,
                              const lbfgsfloatval_t gnorm,
                              const lbfgsfloatval_t step,
                              int n,
                              int k,
                              int ls){
    
    return reinterpret_cast<WavOpticalFlow*>(instance)->progress(x, g, fx, xnorm, gnorm, step, n, k, ls);
};

//=================
//-----------period
//=================
inline int WavOpticalFlow::period(int ind,
                                  int max){
    return (ind<0)?((max + ind%max)%max):ind%max;
};
